// cert.c
// pktlab certificate functions that do not rely on external ssl lib
//

#include <pktlab.h>

#include <assert.h>
#include <ctype.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "private.h"

//
// INTERNAL CONSTANTS
//

#define PKTLAB_URILEN_MIN 12

//
// INTERNAL TYPE DEFINITIONS
//

struct cert_type_name_id_entry {
    const char * name;
    int id;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_name_char(int c);
static char * strchr_advlen (
    char * str, int c, uint_fast32_t * advlen);

//
// EXPORTED FUNCTION DEFINITIONS
//

int pktlab_parse_uri (
    struct pktlab_uri * restrict uri_info,
    const void * restrict ptr, uint_fast32_t len)
{
    char * uri;
    uint_fast32_t i;
    char * cursor;
    char * host;
    uint_fast32_t hostlen;
    char * type;
    uint_fast32_t typelen;
    uint_fast32_t actualhostlen;
    char * port;
    uint_fast32_t portlen;
    unsigned long tmpport;

    trace("%s(urilen:%u)", __func__, (unsigned) len);

    if (uri_info == NULL || ptr == NULL || len > SIZE_MAX-1) {
        debug("%s(uri_info:%p,uri:%p,urilen:%u): bad argument",
            __func__, uri_info, ptr, (unsigned) len);
        return -1;
    }

    uri = safe_malloc(len+1);
    memcpy(uri, ptr, len);
    uri[len] = '\0';

    cursor = uri;
    if (len < PKTLAB_URILEN_MIN)
        goto malformed;

    // check charset
    for (i = 0; i < len; ++i)
        if (!is_name_char(cursor[i]))
            goto malformed;

    // check scheme
    if (strncmp(cursor, "pktlab://", 9) != 0)
        goto malformed;

    // get host
    cursor += 9;
    host = cursor;
    cursor = strchr_advlen(host, '/', &hostlen);
    if (cursor == NULL)
        goto malformed;
    if (hostlen == 0)
        goto malformed;

    // get host
    cursor += 1;
    type = cursor;
    cursor = strchr_advlen(type, '/', &typelen);
    if (cursor == NULL)
        goto malformed;
    if (typelen != 3 && typelen != 6) // not "exp" or "broker"
        goto malformed;

    // (cursor-type) overflow prevention can be done another way,
    // i.e. using sth like type+3 != cursor etc.,
    // but I was too ignorant to come up with it when I wrote the code

    if (hostlen > UINT_FAST32_MAX-11-typelen)
        goto malformed;

    if (len != hostlen+typelen+11)
        goto malformed;

    // get port
    port = NULL;
    cursor = strchr_advlen(host, ':', &actualhostlen);
    if (cursor != NULL) {
        if (actualhostlen > hostlen) // == is not possible
            goto malformed;

        port = cursor+1;
        portlen = hostlen-actualhostlen-1;
        hostlen = actualhostlen;
        if (portlen == 0 || portlen > 5 || hostlen == 0) // port in [0,65535]
            goto malformed;

        for (i = 0; i < portlen; ++i)
            if (!isdigit(port[i]))
                goto malformed;
    }

    tmpport = 0;
    if (port != NULL) {
        tmpport = strtoul(port, NULL, 10);
        if (tmpport > UINT16_MAX)
            goto malformed;
    }

    // get path
    if (typelen == 3 && strncmp(type, "exp", 3) == 0) {
        uri_info->type = EXP_URI_TYPE;
    } else if (typelen == 6 && strncmp(type, "broker", 6) == 0) {
        uri_info->type = BROKER_URI_TYPE;
    } else {
        goto malformed;
    }

    uri_info->uri.host = ptr + (host-uri);
    uri_info->uri.hostlen = hostlen;
    uri_info->uri.port = pktlab_hton16(tmpport);
    free(uri);
    return 0;

malformed:
    debug("%s(urilen:%u): URI malformed", __func__, (unsigned) len);
    uri_info->type = MALFORMED_URI_TYPE;
    uri_info->malformed.ptr = ptr;
    uri_info->malformed.len = len;
    free(uri);
    return -2;
}

const char * pktlab_cert_type_name(int cert_type) {
    const char * names[] = {
        [PKTLAB_CERT_SUBCMD]  = PKTLAB_CERTTYPE_STR_SUBCMD,
        [PKTLAB_CERT_PUBCMD]  = PKTLAB_CERTTYPE_STR_PUBCMD,
        [PKTLAB_CERT_EXPPRIV] = PKTLAB_CERTTYPE_STR_EXPPRIV,
        [PKTLAB_CERT_DELPRIV] = PKTLAB_CERTTYPE_STR_DELPRIV,
        [PKTLAB_CERT_AGENT]   = PKTLAB_CERTTYPE_STR_AGENT,
        [PKTLAB_CERT_UNKNOWN] = PKTLAB_CERTTYPE_STR_UNKNOWN
    };

    trace("%s(cert_type:%d)", __func__, cert_type);

    if (cert_type < 0 ||
        cert_type >= sizeof(names)/sizeof(char *))
        return NULL;

    return names[cert_type]; // can also be NULL
}

int pktlab_cert_type_name_to_id(const char * name) {
    const struct cert_type_name_id_entry mapping[] = {
        { PKTLAB_CERTTYPE_STR_SUBCMD,  PKTLAB_CERT_SUBCMD  },
        { PKTLAB_CERTTYPE_STR_PUBCMD,  PKTLAB_CERT_PUBCMD  },
        { PKTLAB_CERTTYPE_STR_EXPPRIV, PKTLAB_CERT_EXPPRIV },
        { PKTLAB_CERTTYPE_STR_DELPRIV, PKTLAB_CERT_DELPRIV },
        { PKTLAB_CERTTYPE_STR_AGENT,   PKTLAB_CERT_AGENT   }
    };
    size_t i, maplen;

    trace("%s(name:%s)", __func__, name);

    if (name == NULL)
        return PKTLAB_CERT_UNKNOWN;

    maplen = sizeof(mapping)/sizeof(*mapping);
    for (i = 0; i < maplen; ++i)
        if (strcmp(mapping[i].name, name) == 0)
            return mapping[i].id;

    return PKTLAB_CERT_UNKNOWN;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_name_char(int c) {
    if (isalnum(c) || c == '.' || c == '/' || c == ':')
        return true;
    return false;
}

// strchr with advlen argument that returns the advanced length
// advlen should be ignored when returning NULL
char * strchr_advlen(char * str, int c, uint_fast32_t * advlen) {
    uint_fast32_t i;

    assert(str != NULL && advlen != NULL);

    *advlen = 0;
    for (i = 0; str[i] != '\0'; ++i) {
        if (str[i] == c)
            return &str[i];

        if (i == UINT_FAST32_MAX) {
            debug("%s(...): string index wrap around", __func__);
            return NULL;
        }

        ++*advlen;
    }

    return NULL;
}