// cert_openssl.c
//

#include <pktlab.h>

#include <assert.h>
#include <limits.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#include "private.h"

#include "cert_openssl.h"
#include "key_openssl.h"
#include "util_openssl.h"

#include <openssl/asn1.h>
#include <openssl/asn1t.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/obj_mac.h>
#include <openssl/objects.h>
#include <openssl/pem.h>
#include <openssl/safestack.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

//
// INTERNAL CONSTANTS
//

#define SID_DERLEN 68

#define SET_RET_N_GOTO(v,l) do { ret = (v); goto l; } while (0)

enum rst_type {
    RST_MUST_SUCCESS,
    RST_MUST_NOT_FOUND,
    RST_EITHER_WORKS
};

//
// INTERNAL TYPE DEFINITIONS
//

struct name_field_entry {
    const char * name;
    bool * field;
};

//
// INTERNAL FUNCTION DECLARATIONS
//

static bool is_valid_rst(int rst, int rst_type);

static int ASN1_TIME2pktlab_time (
    const ASN1_TIME * at, pktlab_time_t * pt);

static int get_ext_bytes (
    const X509 * x509,
    int target_nid, bool * critical,
    uint8_t ** ptr, uint_fast32_t * len);

static int set_rst_type (
    int cert_type,
    int * rst_type_cert_desc, int * rst_type_filters,
    int * rst_type_monitors, int * rst_type_priority,
    int * rst_type_del_type);

static int safe_comp_strmem(const char * str, uint_fast32_t * totalmem);
static int safe_comp_digestmem (
    uint_fast32_t dcnt, uint_fast32_t * totalmem);

static int set_info_asciistr (
    const char * asciistr, char ** info_field,
    uint8_t ** info_content_ptr);
static int set_info_digests (
    const uint8_t * digests,
    uint_fast32_t dcnt, uint_fast32_t dlen,
    uint8_t ** info_field_ptr,
    uint_fast32_t * info_field_len,
    uint8_t ** info_content_ptr);

static int create_info (
    int cert_type, const uint8_t * publickey,
    const uint8_t * serial, uint_fast32_t seriallen,
    pktlab_time_t not_before, pktlab_time_t not_after,
    const uint8_t * aid, const uint8_t * sid, bool ca,
    long pathlen, const char * cert_desc, const char * priority,
    const uint8_t * filter_digests, uint_fast32_t ftcnt,
    const uint8_t * monitor_digests, uint_fast32_t mtcnt,
    uint8_t del_type, uint8_t * mem);

static bool is_good_del_type(uint8_t del_type);

static bool is_same_subjkey (
    const uint8_t * subjk,
    struct pktlab_certificate * cert);
static int set_subable_channel (
    struct pktlab_certificate * exppriv_cert,
    struct pktlab_publickey * eopk,
    uint8_t * subable_channel);
static int get_info_cnt (
    const struct pktlab_cert_info * info,
    uint_fast32_t * fno, uint_fast32_t * mno,
    uint_fast32_t * pno, uint_fast32_t * plen);
static int get_info_fields (
    const struct pktlab_cert_info * info,
    uint8_t ** ftdigests, uint8_t ** mtdigests, char ** priority);
static int set_digest_n_priority (
    struct pktlab_certificate ** verified_cert_ls,
    uint_fast32_t verified_certnum,
    uint8_t ** ftdigests, uint_fast32_t * ftcnt,
    uint8_t ** mtdigests, uint_fast32_t * mtcnt,
    char *** priorities, uint_fast32_t * pcnt);

//
// EXPORTED FUNCTION DEFINITIONS
//

// pktlab.h functions

struct pktlab_certificate * pktlab_create_cert(void) {
    struct pktlab_certificate * ret;

    trace("%s()", __func__);

    ret = safe_malloc(sizeof(struct pktlab_certificate));
    ret->x509 = NULL;
    return ret;
}

int pktlab_load_cert_pem (
    const char * pem, uint_fast32_t pemlen,
    struct pktlab_certificate * cert)
{
    int ret;
    BIO * bio;

    trace("%s(pemlen:%u)", __func__, (unsigned) pemlen);

    if (pem == NULL || pemlen == 0 || cert == NULL || pemlen > INT_MAX) {
        debug("%s(pem:%p,pemlen:%u,cert:%p): bad argument",
            __func__, pem, (unsigned) pemlen, cert);
        return -1;
    }

    ret = 0;
    bio = BIO_new_mem_buf(pem, pemlen);
    cert->x509 = PEM_read_bio_X509(bio, NULL, NULL, NULL);
    if (cert->x509 == NULL) {
        error_debug_msg("PEM_read_bio_X509");
        ret = -1;
    }

    BIO_free_all(bio);
    return ret;
}

int pktlab_load_cert_der (
    const uint8_t * der, uint_fast32_t derlen,
    struct pktlab_certificate * cert)
{
    int ret;
    const uint8_t * ptr;

    trace("%s(derlen:%u)", __func__, (unsigned) derlen);

    if (der == NULL || derlen == 0 || cert == NULL || derlen > LONG_MAX) {
        debug("%s(der:%p,derlen:%u,cert:%p): bad argument",
            __func__, der, (unsigned) derlen, cert);
        return -1;
    }

    ret = 0;
    ptr = der;
    cert->x509 = d2i_X509(NULL, &ptr, derlen);
    if (cert->x509 == NULL) {
        error_debug_msg("d2i_X509");
        ret = -1;
    }

    return ret;
}

int pktlab_get_cert_der (
    const struct pktlab_certificate * cert,
    void ** der, uint_fast32_t * derlen)
{
    uint8_t * ptr = NULL;
    int rst;

    trace("%s(...)", __func__);

    if (cert == NULL || cert->x509 == NULL ||
        der == NULL || derlen == NULL)
        return -1;

    rst = i2d_X509(cert->x509, &ptr);
    if (rst < 0)
        return -1;

    if (rst > SIZE_MAX || rst > UINT_FAST32_MAX) {
        OPENSSL_free(ptr);
        return -1;
    }

    *derlen = rst;
    *der = safe_malloc(rst);
    memcpy(*der, ptr, rst);
    OPENSSL_free(ptr);
    return 0;
}

int pktlab_decode_certificate (
    const struct pktlab_certificate * cert,
    const struct pktlab_publickey * signkey,
    struct pktlab_cert_info ** ptr)
{
    int ret = PKTLAB_DECODE_UNKNOWN_FAULT;

    // always present
    int cert_type;
    uint8_t * serial;
    uint_fast32_t seriallen;
    pktlab_time_t not_before, not_after;
    uint8_t publickey[PKTLAB_PUBLICKEY_LEN];
    uint8_t comp_sid[PKTLAB_KEYID_LEN];
    uint8_t aid[PKTLAB_KEYID_LEN];
    uint8_t sid[PKTLAB_KEYID_LEN];
    int rst;
    bool ca;
    uint_fast32_t infomem;

    // optional fields
    long pathlen;
    char * cert_desc;
    char * priority;
    uint8_t * filter_digests;
    uint_fast32_t ftcnt;
    uint8_t * monitor_digests;
    uint_fast32_t mtcnt;
    uint8_t del_type;

    int rst_type_cert_desc, rst_type_filters, \
        rst_type_monitors, rst_type_priority, \
        rst_type_del_type;

    trace("%s(...)", __func__);

    if (cert == NULL || ptr == NULL || cert->x509 == NULL) {
        debug("%s(cert:%p,ptr:%p): bad argument", __func__,
            cert, ptr);
        SET_RET_N_GOTO(PKTLAB_DECODE_INVAL, fail);
    }

    if (try_init_openssl() < 0) {
        debug("%s(...): try_init_openssl failed", __func__);
        goto fail;
    }

    // checked field values
    if (!is_ed25519_signalg(cert->x509)) {
        debug("%s(...): Cert unsigned signature algorithm is not Ed25519", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SIGALG, fail);
    }

    if (signkey != NULL && // verify sig optional
        pktlab_verify_certificate_signature(cert, signkey) < 0) {
        debug("%s(...): Cert bad signature", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SIGNATURE, fail);
    }

    if (!is_ver_3(cert->x509)) {
        debug("%s(...): Cert is not X509 v3", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_X509_VER, fail);
    }

    if (!is_ed25519_signature(cert->x509)) {
        debug("%s(...): Cert signed signature algorithm is not Ed25519", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SIGALG, fail);
    }

    if (are_do_not_use_fields_used(cert->x509)) {
        debug("%s(...): Cert using unused field", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_USING_UNUSED_FIELD, fail);
    }

    // get always present fields
    cert_type = get_cert_type(cert->x509);
    if (cert_type < 0) {
        debug("%s(...): get_cert_type failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_CERT_TYPE, fail);
    }

    if (!has_valid_keyusage(cert->x509, cert_type)) { // we need cert_type for this, so delay it until here
        debug("%s(...): Cert bad keyusage", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_KEYUSAGE, fail);
    }

    if (get_serial(cert->x509, &serial, &seriallen) < 0) {
        debug("%s(...): get_serial failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SERIAL, fail);
    }

    if (get_validity(cert->x509, &not_before, &not_after) < 0) {
        debug("%s(...): get_validity failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_VALIDITY, fail_serial);
    }

    if (get_subj_pubkey(cert->x509, publickey) < 0) {
        debug("%s(...): get_subj_pubkey failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SUBJ_PUBKEY, fail_serial);
    }

    if (get_aid(cert->x509, aid, sizeof(aid)) < 0) {
        debug("%s(...): get_aid failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_AID, fail_serial);
    }

    if (get_sid(cert->x509, sid, sizeof(sid)) < 0) {
        debug("%s(...): get_sid failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SID, fail_serial);
    }

    if (comp_keyid(publickey, comp_sid) < 0) {
        debug("%s(...): comp_sid failed", __func__);
        goto fail_serial;
    }

    if (memcmp(sid, comp_sid, sizeof(sid)) != 0) { // ret bad sid for mismatch as well
        debug("%s(...): sid does not match", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_SID, fail_serial);
    }

    rst = get_basic_constraints(cert->x509, &ca, &pathlen); // bad bc for read fail and bad ca pathlen pair
    if (rst < 0) {
        debug("%s(...): get_basic_constraints failed", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_BC, fail_serial);
    }
    if (!is_valid_ca_pathlen(cert_type, ca, pathlen)) {
        debug("%s(...): bad ca & pathlen combination: cert_type:%d, ca:%d, pathlen:%d",
            __func__, cert_type, (int) ca, (int) pathlen);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_BC, fail_serial);
    }
    if (pathlen < 0) {
        pathlen = -1;
    } else if (pathlen > INT_FAST32_MAX)
        pathlen = INT_FAST32_MAX;

    // get and check optional fields
    if (set_rst_type(
            cert_type,
            &rst_type_cert_desc, &rst_type_filters,
            &rst_type_monitors, &rst_type_priority,
            &rst_type_del_type) < 0) {
        debug("%s(...): set_rst_type failed", __func__);
        goto fail_serial;
    }

    if (!is_valid_rst(
            get_cert_desc(cert->x509, &cert_desc),
            rst_type_cert_desc)) { // free cert_desc when failed is fine
        debug("%s(...): invalid cert_desc value", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_CERT_DESC, fail_cert_desc);
    }

    if (!is_valid_rst(
            get_priority(cert->x509, &priority),
            rst_type_priority)) { // free priority when failed is fine
        debug("%s(...): invalid priority value", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_PRIORITY, fail_priority);
    }

    if (!is_valid_rst(
            get_filter_digests(cert->x509, &filter_digests, &ftcnt),
            rst_type_filters)) { // free filter_digests when failed is fine
        debug("%s(...): invalid filter digest value", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_FILTER_DIGESTS, fail_filter);
    }

    if (!is_valid_rst(
            get_monitor_digests(cert->x509, &monitor_digests, &mtcnt),
            rst_type_monitors)) { // free monitor_digests when failed is fine
        debug("%s(...): invalid monitor digest value", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_MONITOR_DIGESTS, fail_monitor);
    }

    if (!is_valid_rst(
            get_del_type(cert->x509, &del_type),
            rst_type_del_type)) {
        debug("%s(...): invalid delegation type value", __func__);
        SET_RET_N_GOTO(PKTLAB_DECODE_BAD_DEL_TYPE, fail_monitor);
    }

    // construct returned structure with retrieved value
    // ret unknown fault for overflow and failure to create info struct
    infomem = sizeof(struct pktlab_cert_info);

    if (infomem > UINT_FAST32_MAX-seriallen) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }
    infomem += seriallen;

    if (cert_desc != NULL && safe_comp_strmem(cert_desc, &infomem) < 0) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }

    if (priority != NULL && safe_comp_strmem(priority, &infomem) < 0) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }

    if (filter_digests != NULL &&
        safe_comp_digestmem(ftcnt, &infomem) < 0) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }

    if (monitor_digests != NULL &&
        safe_comp_digestmem(mtcnt, &infomem) < 0) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }

    if (infomem > SIZE_MAX) {
        debug("%s(...): infomem overflow", __func__);
        goto fail_monitor;
    }

    *ptr = safe_malloc(infomem);
    if (create_info(
            cert_type, publickey, serial, seriallen,
            not_before, not_after, aid, sid, ca,
            pathlen, cert_desc, priority,
            filter_digests, ftcnt,
            monitor_digests, mtcnt,
            del_type, (uint8_t *) *ptr) < 0) {
        free(*ptr);
        debug("%s(...): create_info failed", __func__);
        goto fail_monitor;
    }

    ret = PKTLAB_DECODE_SUCCESS; // successful

fail_monitor:
    free(monitor_digests);
fail_filter:
    free(filter_digests);
fail_priority:
    free(priority);
fail_cert_desc:
    free(cert_desc);
fail_serial:
    free(serial);
fail:
    return ret;
}

int pktlab_verify_certificate_signature (
    const struct pktlab_certificate * cert,
    const struct pktlab_publickey * signkey)
{
    struct pktlab_publickey * signkey_dup;

    trace("%s(...)", __func__);

    if (cert == NULL || signkey == NULL)
        return -1;

    if (dup_key(signkey, &signkey_dup) < 0)
        return -1;

    if (!has_valid_signature(cert->x509, signkey_dup)) {
        pktlab_cleanup_publickey(signkey_dup);
        return -1;
    }

    pktlab_cleanup_publickey(signkey_dup);
    return 0;
}

int pktlab_cleanup_cert(struct pktlab_certificate * cert) {

    trace("%s(...)", __func__);

    if (cert == NULL) {
        debug("%s(cert:%p): bad argument", __func__, cert);
        return -1;
    }

    if (cert->x509 != NULL)
        X509_free(cert->x509);

    free(cert);
    return 0;
}

int pktlab_get_agent_key (
    const struct pktlab_certificate * agent_cert,
    uint8_t * agent_pubkey)
{
    struct pktlab_cert_info * info;
    int rst, ret = -128;

    trace("%s(...)", __func__);

    if (agent_cert == NULL || agent_pubkey == NULL)
        return -1;

    rst = pktlab_decode_certificate(
        agent_cert, NULL, &info);
    if (rst != PKTLAB_DECODE_SUCCESS)
        return -2;

    if (info->cert_type != PKTLAB_CERT_AGENT) {
        ret = -2;
        goto fail_info;
    }

    memcpy(agent_pubkey, info->publickey, sizeof(info->publickey));
    ret = 0;

fail_info:
    free(info);
    return ret;
}

int pktlab_parse_endpoint_rep_cert_chain (
    struct pktlab_certificate ** cert_ls,
    uint_fast32_t certnum, const uint8_t * endpoint_key,
    struct pktlab_publickey * root_eop_key,
    uint8_t ** subable_channels, uint_fast32_t * channelnum)
{
    struct pktlab_publickey ** root_ls;
    uint_fast32_t rootnum;
    struct pktlab_certificate ** verified_cert_ls;
    uint_fast32_t verified_certnum, totalmem, i;
    struct pktlab_cert_info * info;
    uint8_t * channels;
    int ret, rst;

    trace("%s(certnum:%u)", __func__, (unsigned) certnum);

    if (cert_ls == NULL || certnum == 0 ||
        subable_channels == NULL || channelnum == NULL)
        return PKTLAB_PARSE_CHAIN_INVAL;

    root_ls = NULL;
    rootnum = 0;
    if (root_eop_key != NULL) {
        root_ls = &root_eop_key;
        rootnum = 1;
    }

    rst = pktlab_verify_cert_chain(
        PKTLAB_CONTROLLER_AUTH_ENDPOINT,
        cert_ls, root_ls, &verified_cert_ls,
        certnum, rootnum, &verified_certnum, NULL);
    debug("pktlab_verify_cert_chain returned %d", rst);
    if (rst != PKTLAB_VERIFY_SUCCESS)
        return PKTLAB_PARSE_CHAIN_BAD_CHAIN;

    assert(verified_certnum >= 1); // at least one subcmd
    ret = PKTLAB_PARSE_CHAIN_UNKNOWN_FAULT;

    if (!is_same_subjkey(
            endpoint_key, verified_cert_ls[0])) {
        ret = PKTLAB_PARSE_CHAIN_BAD_KEY;
        goto fail_verified_certs;
    }

    totalmem = verified_certnum*PKTLAB_KEYID_LEN;
    if (totalmem/PKTLAB_KEYID_LEN != verified_certnum ||
        totalmem > SIZE_MAX)
        goto fail_verified_certs;
    channels = safe_malloc(totalmem);

    for (i = 0; i < verified_certnum; ++i) {
        rst = pktlab_decode_certificate(
                verified_cert_ls[i],
                NULL, &info);
        assert(
            rst == PKTLAB_DECODE_SUCCESS &&
            (info->cert_type == PKTLAB_CERT_SUBCMD ||
             info->cert_type == PKTLAB_CERT_DELPRIV));

        memcpy(
            channels+i*PKTLAB_KEYID_LEN, info->authority_key_id,
            sizeof(info->authority_key_id));
        free(info);
    }

    free(verified_cert_ls);
    *subable_channels = channels;
    *channelnum = verified_certnum;
    return PKTLAB_PARSE_CHAIN_SUCCESS;

fail_verified_certs:
    free(verified_cert_ls);
    return ret;
}

int pktlab_parse_controller_exp_cert_chain (
    struct pktlab_certificate ** cert_ls,
    uint_fast32_t certnum,
    const uint8_t * controller_key,
    struct pktlab_publickey * trusted_eop_key,
    uint8_t ** ftdigests, uint_fast32_t * ftcnt,
    uint8_t ** mtdigests, uint_fast32_t * mtcnt,
    char *** priorities, uint_fast32_t * pcnt,
    uint8_t * subable_channel)
{
    struct pktlab_certificate ** verified_cert_ls;
    uint_fast32_t verified_certnum;
    int ret;

    trace("%s(certnum:%u)",
        __func__, (unsigned) certnum);

    if (cert_ls == NULL ||
        certnum == 0 ||
        (ftdigests == NULL) != (ftcnt == NULL) ||
        (mtdigests == NULL) != (mtcnt == NULL) ||
        (priorities == NULL) != (pcnt == NULL) ||
        subable_channel == NULL)
        return PKTLAB_PARSE_CHAIN_INVAL;

    if (pktlab_verify_cert_chain(
            PKTLAB_ENDPOINT_AUTH_CONTROLLER,
            cert_ls, NULL, &verified_cert_ls,
            certnum, 0, &verified_certnum, NULL) != PKTLAB_VERIFY_SUCCESS)
        return PKTLAB_PARSE_CHAIN_BAD_CHAIN;

    assert(verified_certnum >= 2);

    if (!is_same_subjkey(
            controller_key, verified_cert_ls[0])) {
        ret = PKTLAB_PARSE_CHAIN_BAD_KEY;
        goto fail_verified_certs;
    }

    if (set_subable_channel(
            verified_cert_ls[verified_certnum-1],
            trusted_eop_key, subable_channel) < 0) {
        ret = PKTLAB_PARSE_CHAIN_BAD_KEY;
        goto fail_verified_certs;
    }

    if (set_digest_n_priority(
            verified_cert_ls, verified_certnum,
            ftdigests, ftcnt, mtdigests, mtcnt,
            priorities, pcnt) < 0) {
        ret = PKTLAB_PARSE_CHAIN_UNKNOWN_FAULT;
        goto fail_verified_certs;
    }

    ret = PKTLAB_PARSE_CHAIN_SUCCESS;

fail_verified_certs:
    free(verified_cert_ls);
    return ret;
}

// cert_openssl.h functions

// This checks for the unsigned one
bool is_ed25519_signalg(const X509 * x509) {
    const ASN1_BIT_STRING * sig;
    const X509_ALGOR * alg;

    assert(x509 != NULL);

    X509_get0_signature(&sig, &alg, x509); // will this do the openssl error queue stuff? no idea
    if (alg == NULL || alg->algorithm == NULL) {
        debug("%s(...): X509_get0_signature failed", __func__);
        return false;
    }

    return (OBJ_obj2nid(alg->algorithm) == NID_ED25519);
}

// Note we do not check whether the aid matches the provided key.
bool has_valid_signature(X509 * x509, struct pktlab_publickey * signkey) {
    int ret;

    assert(x509 != NULL && signkey != NULL);

    ret = X509_verify(x509, signkey->key);
    if (ret == -1) // not validation error
        error_debug_msg("X509_verify");

    return (ret == 1);
}

// This checks for the signed one
bool is_ed25519_signature(const X509 * x509) {
    const X509_ALGOR * alg;

    assert(x509 != NULL);

    alg = X509_get0_tbs_sigalg(x509);
    if (alg == NULL || alg->algorithm == NULL) {
        debug("%s(...): X509_get0_tbs_sigalg failed", __func__);
        return false;
    }

    return (OBJ_obj2nid(alg->algorithm) == NID_ED25519);
}

bool is_ver_3(const X509 * x509) {
    assert(x509 != NULL);
    return (X509_get_version(x509)+1 == 3);
}

bool has_valid_keyusage(X509 * x509, int cert_type) {
    uint32_t keyusage;

    assert(x509 != NULL);

    keyusage = X509_get_key_usage(x509);
    if (keyusage == UINT32_MAX) {
        debug("%s(cert_type:%d): X509_get_key_usage keyusage field absent", __func__, cert_type);
        return false;
    }

    switch(cert_type) {
    case PKTLAB_CERT_EXPPRIV:
    case PKTLAB_CERT_DELPRIV:
        return (keyusage == KU_KEY_CERT_SIGN || \
                keyusage == (KU_KEY_CERT_SIGN|KU_CRL_SIGN));
    case PKTLAB_CERT_SUBCMD:
    case PKTLAB_CERT_PUBCMD:
    case PKTLAB_CERT_AGENT:
        return (keyusage == KU_DIGITAL_SIGNATURE);
    default:
        debug("%s(cert_type:%d): unknown cert_type", __func__, cert_type);
    }

    return false;
}

bool is_valid_ca_pathlen(int cert_type, bool ca, long pathlen) {
    switch(cert_type) {
    case PKTLAB_CERT_EXPPRIV:
    case PKTLAB_CERT_DELPRIV:
        return ca; // pathlen can be arbitrary
    case PKTLAB_CERT_SUBCMD:
    case PKTLAB_CERT_PUBCMD:
    case PKTLAB_CERT_AGENT:
        return pathlen < 0 && !ca;
    default:
        debug("%s(cert_type:%d): unknown cert_type", __func__, cert_type);
    }

    return false;
}

bool are_do_not_use_fields_used(const X509 * x509) {
    const ASN1_BIT_STRING * iuid;
    const ASN1_BIT_STRING * suid;
    const STACK_OF(X509_EXTENSION) * exts;
    X509_EXTENSION * ext;
    ASN1_OBJECT * obj;
    int i, nid, extcnt;
    uint_fast32_t allowedcnt, j;
    bool allowed;
    const int allowed_nids[] = {
        NID_authority_key_identifier, // Authority Key Identifier
        NID_subject_key_identifier,   // Subject Key Identifier
        NID_key_usage,                // Key Usage
        NID_basic_constraints,        // Basic Constraints
        NID_crl_distribution_points,  // CRL Distribution Points
        NID_freshest_crl,             // Freshest CRL
        NID_info_access,              // Authority Information Access
        NID_sinfo_access,             // Subject Information Access
        _nid_cert_type,               // pktlab Cert Type
        _nid_cert_desc,               // Certificate description
        _nid_filter_digests,          // Filter digests
        _nid_monitor_digests,         // Monitor digests
        _nid_priority,                // Priority
        _nid_del_type                 // Delegation type
    }; // probably should use hash table or binary search for this

    assert(x509 != NULL && _nid_defined);

    // check no unique ID
    X509_get0_uids(x509, &iuid, &suid);
    if (iuid != NULL || suid != NULL) {
        debug("%s(...): unique identifier field(s) present", __func__);
        return true;
    }

    // check only contains allowed extension fields
    exts = X509_get0_extensions(x509);
    if (exts == NULL) {
        debug("%s(...): found no extension field", __func__);
        return true;
    }

    extcnt = sk_X509_EXTENSION_num(exts);
    allowedcnt = sizeof(allowed_nids)/sizeof(*allowed_nids);
    for (i = 0; i < extcnt; ++i) {
        ext = sk_X509_EXTENSION_value(exts, i);
        if (ext == NULL) {
            debug("%s(...): index (%d/%d) out of range", __func__, i+1, extcnt);
            return true;
        }

        obj = X509_EXTENSION_get_object(ext);
        if (obj == NULL) {
            debug("%s(...): getting extension OID object failed", __func__);
            return true;
        }

        nid = OBJ_obj2nid(obj);
        allowed = false;
        for (j = 0; j < allowedcnt; ++j) {
            if (allowed_nids[j] == nid) {
                allowed = true;
                break;
            }
        }

        if (!allowed) {
            debug("%s(...): unused extension (%d) used", __func__, nid);
            return true;
        }
    }

    return false;
}

// get field value

// Passes back allocated memory, need to be freed afterwards.
int get_serial (
    const X509 * x509,
    uint8_t ** buf, uint_fast32_t * len) // serial in big endian bytes
{
    const ASN1_INTEGER * serial;
    BIGNUM * bn_serial;
    int needlen;

    assert(x509 != NULL && buf != NULL && len != NULL);

    serial = X509_get0_serialNumber(x509);
    if (serial == NULL) {
        debug("%s(...): X509_get0_serialNumber failed", __func__);
        goto fail;
    }

    bn_serial = ASN1_INTEGER_to_BN(serial, NULL);
    if (bn_serial == NULL) {
        debug("%s(...): ASN1_INTEGER_to_BN failed", __func__);
        goto fail;
    }

    needlen = BN_num_bytes(bn_serial);
    if (needlen <= 0 || needlen > UINT32_MAX || needlen > SIZE_MAX) {
        debug("%s(...): BN_num_bytes returned invalid value (%d)",
            __func__, needlen);
        goto fail_free;
    }

    *buf = safe_malloc(needlen);
    *len = needlen;

    assert(BN_bn2bin(bn_serial, *buf) == needlen);

    BN_free(bn_serial);
    return 0;

fail_free:
    BN_free(bn_serial);
fail:
    *buf = NULL;
    *len = 0;
    return -1;
}

int get_aid(const X509 * x509, void * buf, uint_fast32_t buflen) {
    AUTHORITY_KEYID * aid;
    int crit;
    const uint8_t * asn1_bytes;
    int len;

    assert(x509 != NULL && buf != NULL && buflen >= PKTLAB_KEYID_LEN);

    aid = X509_get_ext_d2i(x509, NID_authority_key_identifier, &crit, NULL);
    if (aid == NULL) {
        debug("%s(...): X509_get_ext_d2i(NID_authority_key_identifier) failed",
            __func__);
        return -1;
    }
    if (crit != 0 || aid->keyid == NULL) {
        debug("%s(...): X509_get_ext_d2i returned invalid critical (%d) or aid",
            __func__, crit);
        AUTHORITY_KEYID_free(aid);
        return -1;
    }

    asn1_bytes = ASN1_STRING_get0_data(aid->keyid);
    if (asn1_bytes == NULL) {
        debug("%s(...): ASN1_STRING_get0_data failed", __func__);
        AUTHORITY_KEYID_free(aid);
        return -1;
    }

    len = ASN1_STRING_length(aid->keyid);
    if (len != PKTLAB_KEYID_LEN) { // should prevent weird external input, as we don't want to parse asn1 seq
        debug("%s(...): Got bad AID len (%d)", __func__, len);
        AUTHORITY_KEYID_free(aid);
        return -1;
    }

    memcpy(buf, asn1_bytes, PKTLAB_KEYID_LEN);

    AUTHORITY_KEYID_free(aid);
    return 0;
}

int get_sid(const X509 * x509, void * buf, uint_fast32_t buflen) {
    bool critical;
    uint8_t * ptr;
    uint_fast32_t len;

    assert(x509 != NULL && buf != NULL && buflen >= PKTLAB_KEYID_LEN);

    if (get_ext_bytes(
            x509, NID_subject_key_identifier,
            &critical, &ptr, &len) < 0) {
        debug("%s(...): get_ext_bytes failed", __func__);
        return -1;
    }

    if (critical || len != PKTLAB_KEYID_LEN) {
        debug("%s(...): invalid SID data retrieved (critical:%d,sidlen:%u)",
            __func__, (int) critical, (unsigned) len);
        free(ptr);
        return -1;
    }

    memcpy(buf, ptr, len);
    free(ptr);
    return 0;
}

int get_validity (
    const X509 * x509,
    pktlab_time_t * not_valid_before,
    pktlab_time_t * not_valid_after)
{
    const ASN1_TIME * not_before;
    const ASN1_TIME * not_after;

    assert(x509 != NULL);

    if (not_valid_before != NULL) {
        not_before = X509_get0_notBefore(x509);
        if (not_before == NULL) {
            debug("%s(...): X509_get0_notBefore failed", __func__);
            return -1;
        }

        if (ASN1_TIME2pktlab_time(not_before, not_valid_before) < 0) {
            debug("%s(...): not_valid_before time translation failed", __func__);
            return -1;
        }
    }

    if (not_valid_after != NULL) {
        not_after = X509_get0_notAfter(x509);
        if (not_after == NULL) {
            debug("%s(...): X509_get0_notAfter failed", __func__);
            return -1;
        }

        if (ASN1_TIME2pktlab_time(not_after, not_valid_after) < 0) {
            debug("%s(...): not_valid_after time translation failed", __func__);
            return -1;
        }
    }

    return 0;
}

int get_subj_pubkey(const X509 * x509, uint8_t * signkey) {
    const EVP_PKEY * k;
    size_t keylen;
    int rst;

    assert(x509 != NULL && signkey != NULL);

    k = X509_get0_pubkey(x509);
    if (k == NULL) {
        debug("%s(...): X509_get0_pubkey failed", __func__);
        return -1;
    }

    rst = EVP_PKEY_get_raw_public_key(k, NULL, &keylen);
    if (rst == 0 || keylen != PKTLAB_PUBLICKEY_LEN) {
        debug("%s(...): EVP_PKEY_get_raw_public_key get keylen failed", __func__);
        return -1;
    }

    rst = EVP_PKEY_get_raw_public_key(k, signkey, &keylen);
    if (rst == 0 || keylen != PKTLAB_PUBLICKEY_LEN) {
        debug("%s(...): EVP_PKEY_get_raw_public_key get key failed", __func__);
        return -1;
    }

    return 0;
}

int get_basic_constraints(X509 * x509, bool * ca, long * pathlen) {
    BASIC_CONSTRAINTS * bc;
    int crit;

    assert(x509 != NULL && ca != NULL && pathlen != NULL);

    bc = X509_get_ext_d2i(x509, NID_basic_constraints, &crit, NULL);
    if (bc == NULL) {
        if (crit == -1)
            return -2; // -2 means ext does not exist

        debug("%s(...): X509_get_ext_d2i(NID_basic_constraints) failed", __func__);
        return -1;
    }
    if (crit != 1) {
        debug("%s(...): X509_get_ext_d2i returned invalid critical (%d)", __func__, crit);
        goto fail;
    }

    *ca = bc->ca != 0;

    *pathlen = X509_get_pathlen(x509);
    if (*pathlen < 0) {
        if (bc->pathlen != NULL) {
            debug("%s(...): X509_get_pathlen failed", __func__);
            goto fail;
        }
    }

    BASIC_CONSTRAINTS_free(bc);
    return 0;

fail:
    BASIC_CONSTRAINTS_free(bc);
    return -1;
}

int get_cert_type(const X509 * x509) {
    char * name;
    int rst, ret;

    assert(x509 != NULL && _nid_defined);

    rst = get_ext_ascii_str(x509, _nid_cert_type, true, &name);
    if (rst < 0)
        return rst; // allow passing -2 for missing ext

    ret = pktlab_cert_type_name_to_id(name);
    return (ret == PKTLAB_CERT_UNKNOWN) ? -1:ret;
}

int get_del_type(const X509 * x509, uint8_t * del_type) {
    int ret;
    bool critical;
    uint8_t * ptr;
    uint_fast32_t len;

    assert(x509 != NULL && del_type != NULL && _nid_defined);

    ret = get_ext_bytes(x509, _nid_del_type, &critical, &ptr, &len);
    if (ret < 0) {
        *del_type = 0;
        return ret;
    }

    if (critical != false) {
        debug("%s(...): invalid critical value (expect 0, got %d)",
            __func__, (int) critical);
        goto fail;
    }

    if (len != sizeof(*del_type)) {
        debug("%s(...): got bad field length: %u",
            __func__, (unsigned) len);
        goto fail;
    }

    if (!is_good_del_type(*ptr)) {
        debug("%s(...): got bad del_type: %u",
            __func__, (unsigned) *ptr);
        goto fail;
    }

    *del_type = *ptr;
    free(ptr);
    return 0;

fail:
    *del_type = 0;
    free(ptr);
    return -1;
}

int get_ext_ascii_str (
    const X509 * x509,
    int nid, bool expectcrit, char ** str)
{
    int ret;
    bool critical;
    uint8_t * ptr;
    uint_fast32_t len;

    assert(x509 != NULL && str != NULL && _nid_defined);

    ret = get_ext_bytes(x509, nid, &critical, &ptr, &len);
    if (ret < 0) {
        *str = NULL;
        return ret;
    }

    if (critical != expectcrit) {
        debug("%s(...): invalid critical value (expect %d, got %d)",
            __func__, (int) expectcrit, (int) critical);
        goto fail;
    }

    if (strlen((char *) ptr) != len) { // check no weird char in ptr
        debug("%s(...): embedded null char in retrieved string", __func__);
        goto fail;
    }

    *str = (char *) ptr;
    return 0;

fail:
    *str = NULL;
    free(ptr);
    return -1;
}

int get_ext_digests_non_critical (
    const X509 * x509,
    int nid, uint_fast32_t digestlen,
    uint8_t ** digests, uint_fast32_t * dcnt)
{
    int ret;
    bool critical;
    uint_fast32_t len;

    assert(x509 != NULL && digests != NULL && dcnt != NULL && _nid_defined);

    ret = get_ext_bytes(x509, nid, &critical, digests, &len);
    if (ret < 0) {
        *digests = NULL;
        *dcnt = 0;
        return ret;
    }

    if (critical == true) {
        debug("%s(...): invalid critical value", __func__);
        goto fail;
    }

    if (len % digestlen != 0) {
        debug("%s(...): digest total length (%u)"
            " not a multiple of digest length (%u)",
            __func__, (unsigned) len, (unsigned) digestlen);
        goto fail;
    }

    *dcnt = len / digestlen;
    return 0;

fail:
    free(*digests);
    *digests = NULL;
    *dcnt = 0;
    return -1;
}

int comp_keyid (
    const uint8_t * rawkey, uint8_t * buf)
{
    struct pktlab_publickey * pkey;

    assert(rawkey != NULL && buf != NULL);

    // comp sid from public key
    pkey = pktlab_create_publickey();
    if (pkey == NULL)
        goto fail;

    if (pktlab_load_publickey_raw(
            rawkey, PKTLAB_PUBLICKEY_LEN, pkey) < 0)
        goto fail_pkey;

    if (pktlab_get_key_identifier(pkey, buf, PKTLAB_KEYID_LEN) < 0)
        goto fail_pkey;

    pktlab_cleanup_publickey(pkey);
    return 0;

fail_pkey:
    pktlab_cleanup_publickey(pkey);
fail:
    return -1;
}

int dup_cert (
    const struct pktlab_certificate * orgcert,
    struct pktlab_certificate ** newcert)
{
    assert(orgcert != NULL && newcert != NULL);

    *newcert = pktlab_create_cert();
    (*newcert)->x509 = X509_dup(orgcert->x509);
    if ((*newcert)->x509 == NULL) {
        pktlab_cleanup_cert(*newcert);
        return -1;
    }

    return 0;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

bool is_valid_rst(int rst, int rst_type) {
    switch (rst_type) {
    case RST_MUST_SUCCESS:
        return (rst == 0);
    case RST_MUST_NOT_FOUND:
        return (rst == -2);
    case RST_EITHER_WORKS:
        return (rst == 0 || rst == -2);
    default:
        break;
    }

    return false;
}

int ASN1_TIME2pktlab_time (
    const ASN1_TIME * at, pktlab_time_t * pt)
{
    struct tm tm;

    assert(at != NULL && pt != NULL);

    if (ASN1_TIME_to_tm(at, &tm) == 0) {
        debug("%s(...): ASN1_TIME_to_tm failed", __func__);
        return -1;
    }

    *pt = pktlab_tm_to_time(&tm);
    return 0;
}

// seach for the extension with nid and get ext byte data
int get_ext_bytes (
    const X509 * x509,
    int target_nid, bool * critical,
    uint8_t ** ptr, uint_fast32_t * len)
{
    const STACK_OF(X509_EXTENSION) * exts;
    int extcnt, i;
    X509_EXTENSION * target;
    X509_EXTENSION * cand;
    ASN1_OBJECT * obj;
    int nid;
    ASN1_STRING * extdata;
    const uint8_t * head;
    int extdatalen;
    const uint8_t * cursor;
    long encodedlen;
    int tag, class;

    assert(
        x509 != NULL && target_nid != NID_undef && critical != NULL &&
        ptr != NULL && len != NULL && _nid_defined);

    // find the target extension
    exts = X509_get0_extensions(x509);
    if (exts == NULL) {
        debug("%s(...): X509_get0_extensions failed", __func__);
        goto fail;
    }

    target = NULL;
    extcnt = sk_X509_EXTENSION_num(exts);
    if (extcnt < 0) { // should never happen
        debug("%s(...): sk_X509_EXTENSION_num failed", __func__);
        goto fail;
    }

    for (i = 0; i < extcnt; ++i) {
        cand = sk_X509_EXTENSION_value(exts, i);
        if (cand == NULL) { // should never happen
            debug("%s(...): sk_X509_EXTENSION_value failed", __func__);
            goto fail;
        }

        obj = X509_EXTENSION_get_object(cand);
        if (obj == NULL) {
            debug("%s(...): X509_EXTENSION_get_object failed", __func__);
            goto fail;
        }

        nid = OBJ_obj2nid(obj);
        if (nid == NID_undef)
            continue;
        if (nid != target_nid)
            continue;

        if (target != NULL) {
            debug("%s(...): duplicated ext (nid:%d) entry found",
                __func__, nid);
            goto fail;
        }
        target = cand;
    }

    if (target == NULL) {
        debug("%s(...): target ext not found", __func__);
        *ptr = NULL;
        *len = 0;
        return -2;
    }

    // get information about the extension
    *critical = X509_EXTENSION_get_critical(target) == 1;

    extdata = X509_EXTENSION_get_data(target);
    if (extdata == NULL) {
        debug("%s(...): X509_EXTENSION_get_data failed", __func__);
        goto fail;
    }

    head = ASN1_STRING_get0_data(extdata);
    if (head == NULL) {
        debug("%s(...): ASN1_STRING_get0_data failed", __func__);
        goto fail;
    }

    extdatalen = ASN1_STRING_length(extdata);
    if (extdatalen <= 0) {
        debug("%s(...): ASN1_STRING_length failed", __func__);
        goto fail;
    }

    cursor = head;
    if (ASN1_get_object(
            &cursor, &encodedlen, &tag, &class, extdatalen) != 0) { // only accept primitive definite
        debug("%s(...): non-primitive-encoding ret value for ASN1_get_object",
            __func__);
        goto fail;
    }
    if (tag != V_ASN1_OCTET_STRING || // only deal with OCTET STRINGs
        class != V_ASN1_UNIVERSAL ||
        cursor+encodedlen != head+extdatalen ||
        encodedlen < 0 || encodedlen >= SIZE_MAX ||
        encodedlen > UINT_FAST32_MAX) {
        debug("%s(...): ASN1_get_object invalid results"
            " (tag:%d,class:%d,encodedlen:%ld)",
            __func__, tag, class, encodedlen);
        goto fail;
    }

    *ptr = safe_malloc(encodedlen+1);
    memcpy(*ptr, cursor, encodedlen);
    (*ptr)[encodedlen] = '\0'; // end null byte for convenience
    *len = encodedlen;

    return 0;

fail:
    *ptr = NULL;
    *len = 0;
    return -1;
}

int set_rst_type (
    int cert_type,
    int * rst_type_cert_desc, int * rst_type_filters,
    int * rst_type_monitors, int * rst_type_priority,
    int * rst_type_del_type)
{
    assert(
        rst_type_cert_desc != NULL &&
        rst_type_filters != NULL &&
        rst_type_monitors != NULL &&
        rst_type_priority != NULL &&
        rst_type_del_type != NULL);

    // default value
    *rst_type_cert_desc = RST_EITHER_WORKS;
    *rst_type_filters   = RST_MUST_NOT_FOUND;
    *rst_type_monitors  = RST_MUST_NOT_FOUND;
    *rst_type_priority  = RST_MUST_NOT_FOUND;
    *rst_type_del_type  = RST_MUST_NOT_FOUND;

    switch (cert_type) {
    case PKTLAB_CERT_SUBCMD:
    case PKTLAB_CERT_AGENT:
        break;
    case PKTLAB_CERT_PUBCMD:
        *rst_type_priority = RST_EITHER_WORKS;
        break;
    case PKTLAB_CERT_DELPRIV:
        *rst_type_del_type = RST_MUST_SUCCESS;
        // fall through
    case PKTLAB_CERT_EXPPRIV:
        *rst_type_filters  = RST_EITHER_WORKS;
        *rst_type_monitors = RST_EITHER_WORKS;
        *rst_type_priority = RST_EITHER_WORKS;
        break;
    case PKTLAB_CERT_UNKNOWN:
    default:
        return -1;
    }

    return 0;
}

int safe_comp_strmem(const char * str, uint_fast32_t * totalmem) {
    size_t len;

    assert(str != NULL && totalmem != NULL);

    len = strlen(str);
    if (len > SIZE_MAX-1 ||
        len+1 > UINT_FAST32_MAX ||
        *totalmem > UINT_FAST32_MAX-(len+1))
        return -1;

    *totalmem += len+1;
    return 0;
}

int safe_comp_digestmem (
    uint_fast32_t dcnt, uint_fast32_t * totalmem)
{
    uint_fast32_t len;

    assert(
        PKTLAB_MONITOR_DIGEST_LEN == PKTLAB_FILTER_DIGEST_LEN &&
        PKTLAB_MONITOR_DIGEST_LEN == PKTLAB_SHA256_DIGEST_LEN);

    len = dcnt * PKTLAB_SHA256_DIGEST_LEN;
    if (len/PKTLAB_SHA256_DIGEST_LEN != dcnt ||
        *totalmem > UINT_FAST32_MAX-len)
        return -1;

    *totalmem += len;
    return 0;
}

int set_info_asciistr (
    const char * asciistr, char ** info_field,
    uint8_t ** info_content_ptr)
{
    assert(info_field != NULL && info_content_ptr != NULL);

    if (asciistr == NULL) {
        *info_field = NULL;
        return 0;
    }

    strcpy((char *) *info_content_ptr, asciistr);
    *info_field = (char *) *info_content_ptr;
    *info_content_ptr += strlen(asciistr)+1;
    return 0;
}

int set_info_digests (
    const uint8_t * digests,
    uint_fast32_t dcnt, uint_fast32_t dlen,
    uint8_t ** info_field_ptr,
    uint_fast32_t * info_field_len,
    uint8_t ** info_content_ptr)
{
    uint_fast32_t totallen;

    assert(
        info_field_ptr != NULL &&
        info_field_len != NULL &&
        info_content_ptr != NULL);

    if (digests == NULL) {
        *info_field_ptr = NULL;
        *info_field_len = 0;
        return 0;
    }

    totallen = dcnt * dlen;
    if (totallen/dlen != dcnt ||
        totallen > SIZE_MAX)
        return -1;

    memcpy(*info_content_ptr, digests, totallen);
    *info_field_ptr = *info_content_ptr;
    *info_field_len = dcnt;
    *info_content_ptr += totallen;

    return 0;
}

int create_info (
    int cert_type, const uint8_t * publickey,
    const uint8_t * serial, uint_fast32_t seriallen,
    pktlab_time_t not_before, pktlab_time_t not_after,
    const uint8_t * aid, const uint8_t * sid, bool ca,
    long pathlen, const char * cert_desc, const char * priority,
    const uint8_t * filter_digests, uint_fast32_t ftcnt,
    const uint8_t * monitor_digests, uint_fast32_t mtcnt,
    uint8_t del_type, uint8_t * mem)
{
    struct pktlab_cert_info * info;
    uint8_t * ptr;

    assert(
        publickey != NULL && serial != NULL &&
        aid != NULL && sid != NULL && mem != NULL); // only check for always present ones
    // cert_desc, priority, filter_digests, monitor_digests may be NULL

    // afterwards, no fail-return checking except overflow/wrap-around & cert_type

    info = (struct pktlab_cert_info *) mem;
    ptr = (uint8_t *) (info+1);

    info->cert_type = cert_type;

    memcpy(ptr, serial, seriallen);
    info->serialno = ptr;
    ptr += seriallen;
    info->serialno_len = seriallen;

    info->valid_not_before = not_before;
    info->valid_not_after = not_after;

    memcpy(info->publickey, publickey, PKTLAB_PUBLICKEY_LEN);
    memcpy(info->authority_key_id, aid, PKTLAB_KEYID_LEN);
    memcpy(info->subject_key_id, sid, PKTLAB_KEYID_LEN);

    if (set_info_asciistr(cert_desc, &info->cert_desc, &ptr) < 0)
        return -1;

    switch (cert_type) {
    case PKTLAB_CERT_EXPPRIV:
        if (set_info_asciistr(priority, &info->exppriv.priority, &ptr) < 0)
            return -1;

        if (set_info_digests(
                filter_digests, ftcnt, PKTLAB_FILTER_DIGEST_LEN,
                &info->exppriv.filter_digests, &info->exppriv.ftcnt, &ptr) < 0)
            return -1;

        if (set_info_digests(
                monitor_digests, mtcnt, PKTLAB_MONITOR_DIGEST_LEN,
                &info->exppriv.monitor_digests, &info->exppriv.mtcnt, &ptr) < 0)
            return -1;

        info->exppriv.pathlen = pathlen;
        break;
    case PKTLAB_CERT_DELPRIV:
        if (set_info_asciistr(priority, &info->delpriv.priority, &ptr) < 0)
            return -1;

        if (set_info_digests(
                filter_digests, ftcnt, PKTLAB_FILTER_DIGEST_LEN,
                &info->delpriv.filter_digests, &info->delpriv.ftcnt, &ptr) < 0)
            return -1;

        if (set_info_digests(
                monitor_digests, mtcnt, PKTLAB_MONITOR_DIGEST_LEN,
                &info->delpriv.monitor_digests, &info->delpriv.mtcnt, &ptr) < 0)
            return -1;

        info->delpriv.pathlen = pathlen;
        info->delpriv.del_type = del_type;
        break;
    case PKTLAB_CERT_PUBCMD:
        if (set_info_asciistr(priority, &info->pubcmd.priority, &ptr) < 0)
            return -1;
        break;
    case PKTLAB_CERT_SUBCMD:
    case PKTLAB_CERT_AGENT:
        break;
    default:
        return -1;
    }

    return 0;
}

bool is_good_del_type(uint8_t del_type) {
    if (del_type == 0 ||
        (del_type &
         ~PKTLAB_DEL_TYPE_EXPPRIV &
         ~PKTLAB_DEL_TYPE_REPPRIV) != 0)
        return false;
    return true;
}

bool is_same_subjkey (
    const uint8_t * subjk,
    struct pktlab_certificate * cert)
{
    struct pktlab_cert_info * info;

    assert(cert != NULL);

    if (subjk == NULL)
        return true; // no checking done when no ck passed

    if (pktlab_decode_certificate(
            cert, NULL, &info) != PKTLAB_DECODE_SUCCESS)
        return false; // failed to decode cert, assume no match

    if (memcmp(subjk, info->publickey, sizeof(info->publickey)) != 0) {
        free(info);
        return false;
    }

    free(info);
    return true;
}

int set_subable_channel (
    struct pktlab_certificate * exppriv_cert,
    struct pktlab_publickey * eopk,
    uint8_t * subable_channel)
{
    struct pktlab_cert_info * info;
    uint8_t keyid[PKTLAB_KEYID_LEN];
    int rst, ret = -1;

    trace("%s(...)", __func__);

    assert(exppriv_cert != NULL);

    if (eopk != NULL &&
        pktlab_get_key_identifier(
            eopk, keyid, sizeof(keyid)) < 0)
        goto fail;

    rst = pktlab_decode_certificate(
        exppriv_cert, eopk, &info);
    if (rst == PKTLAB_DECODE_BAD_SIGNATURE) {
        goto fail;
    } else if (eopk != NULL &&
               rst == PKTLAB_DECODE_SUCCESS &&
               memcmp(
                   info->authority_key_id,
                   keyid, sizeof(keyid)) != 0) {
        goto fail_info;
    }

    assert(
        rst == PKTLAB_DECODE_SUCCESS &&
        info->cert_type == PKTLAB_CERT_EXPPRIV);
    if (subable_channel != NULL)
        memcpy(
            subable_channel, info->authority_key_id,
            sizeof(info->authority_key_id));

    ret = 0;

fail_info:
    free(info);
fail:
    return ret;
}

int get_info_cnt (
    const struct pktlab_cert_info * info,
    uint_fast32_t * fno, uint_fast32_t * mno,
    uint_fast32_t * pno, uint_fast32_t * plen)
{
    size_t len;

    assert(
        info != NULL && fno != NULL &&
        mno != NULL && pno != NULL && plen != NULL);

    switch (info->cert_type) {
    case PKTLAB_CERT_EXPPRIV:
        *fno = info->exppriv.ftcnt;
        *mno = info->exppriv.mtcnt;
        *pno = (info->exppriv.priority == NULL) ? 0:1;
        *plen = 0;
        if (info->exppriv.priority != NULL) {
            len = strlen(info->exppriv.priority);
            if (len == SIZE_MAX ||
                len > UINT_FAST32_MAX-1)
                return -1;
            *plen = len+1;
        }
        break;
    case PKTLAB_CERT_DELPRIV:
        *fno = info->delpriv.ftcnt;
        *mno = info->delpriv.mtcnt;
        *pno = (info->delpriv.priority == NULL) ? 0:1;
        *plen = 0;
        if (info->delpriv.priority != NULL) {
            len = strlen(info->delpriv.priority);
            if (len == SIZE_MAX ||
                len > UINT_FAST32_MAX-1)
                return -1;
            *plen = len+1;
        }
        break;
    default:
        return -1;
    }

    return 0;
}

int get_info_fields (
    const struct pktlab_cert_info * info,
    uint8_t ** ftdigests, uint8_t ** mtdigests, char ** priority)
{
    assert(
        info != NULL && ftdigests != NULL &&
        mtdigests != NULL && priority != NULL);

    switch (info->cert_type) {
    case PKTLAB_CERT_EXPPRIV:
        *ftdigests = info->exppriv.filter_digests;
        *mtdigests = info->exppriv.monitor_digests;
        *priority = info->exppriv.priority;
        break;
    case PKTLAB_CERT_DELPRIV:
        *ftdigests = info->delpriv.filter_digests;
        *mtdigests = info->delpriv.monitor_digests;
        *priority = info->delpriv.priority;
        break;
    default:
        return -1;
    }

    return 0;
}

int set_digest_n_priority (
    struct pktlab_certificate ** verified_cert_ls,
    uint_fast32_t verified_certnum,
    uint8_t ** ftdigests, uint_fast32_t * ftcnt,
    uint8_t ** mtdigests, uint_fast32_t * mtcnt,
    char *** priorities, uint_fast32_t * pcnt)
{
    uint_fast32_t totalmem, i, fno, mno, pno, plen;
    uint_fast32_t totalfno, totalmno, totalpno, totalplen, infonum;
    struct pktlab_cert_info ** info_ls;
    uint8_t * ftds;
    uint8_t * mtds;
    char * p;
    uint8_t * ftdptr;
    uint8_t * mtdptr;
    char ** parrptr = NULL;
    char * psptr = NULL;
    int rst;

    trace("%s(verified_certnum:%u)",
        __func__, (unsigned) verified_certnum);

    assert(
        verified_cert_ls != NULL && verified_certnum != 0 &&
        (ftdigests == NULL) == (ftcnt == NULL) &&
        (mtdigests == NULL) == (mtcnt == NULL) &&
        (priorities == NULL) == (pcnt == NULL));

    totalmem = (verified_certnum-1)*sizeof(*info_ls);
    if (totalmem/sizeof(*info_ls) != verified_certnum-1 ||
        totalmem > SIZE_MAX)
        return -1;
    info_ls = safe_malloc(totalmem);

    // collect memory information
    infonum = totalfno = totalmno = totalpno = totalplen = 0;
    for (i = verified_certnum-1; i > 0; --i) {
        rst = pktlab_decode_certificate(
            verified_cert_ls[i], NULL,
            &info_ls[verified_certnum-1-i]);
        assert(rst == PKTLAB_DECODE_SUCCESS);
        ++infonum;

        rst = get_info_cnt(
            info_ls[verified_certnum-1-i],
            &fno, &mno, &pno, &plen);
        assert(rst >= 0);

        if (fno > UINT_FAST32_MAX-totalfno)
            goto fail;
        totalfno += fno;

        if (mno > UINT_FAST32_MAX-totalmno)
            goto fail;
        totalmno += mno;

        if (pno > UINT_FAST32_MAX-totalpno)
            goto fail;
        totalpno += pno;

        if (plen > UINT_FAST32_MAX-totalplen)
            goto fail;
        totalplen += plen;
    }

    // set filter digests
    if (ftdigests != NULL) {
        *ftdigests = NULL;
        totalmem = totalfno*PKTLAB_FILTER_DIGEST_LEN;
        if (totalmem/PKTLAB_FILTER_DIGEST_LEN != totalfno ||
            totalmem > SIZE_MAX)
            goto fail;
        ftdptr = *ftdigests = safe_malloc(totalmem);
        *ftcnt = 0;
    }

    if (mtdigests != NULL) {
        *mtdigests = NULL;
        totalmem = totalmno*PKTLAB_MONITOR_DIGEST_LEN;
        if (totalmem/PKTLAB_MONITOR_DIGEST_LEN != totalmno ||
            totalmem > SIZE_MAX)
            goto fail_ftd;
        mtdptr = *mtdigests = safe_malloc(totalmem);
        *mtcnt = 0;
    }

    if (priorities != NULL) {
        *priorities = NULL;
        totalmem = sizeof(**priorities)*totalpno;
        if (totalmem/sizeof(**priorities) != totalpno)
            goto fail_mtd;
        if (totalmem > UINT_FAST32_MAX-totalplen ||
            totalmem+totalplen > SIZE_MAX)
            goto fail_mtd;
        totalmem += totalplen;
        parrptr = *priorities = safe_malloc(totalmem);
        psptr = (char *) ((*priorities)+totalpno); // to ptr array end
        *pcnt = 0;
    }

    for (i = 0; i < infonum; ++i) {
        rst = get_info_cnt(
            info_ls[i], &fno, &mno, &pno, &plen);
        assert(rst >= 0);

        rst = get_info_fields(
            info_ls[i], &ftds, &mtds, &p);
        assert(rst >= 0);

        if (ftdigests != NULL && ftds != NULL) {
            memcpy(ftdptr, ftds, fno*PKTLAB_FILTER_DIGEST_LEN);
            ftdptr += fno*PKTLAB_FILTER_DIGEST_LEN;
            *ftcnt += fno;
        }

        if (mtdigests != NULL && mtds != NULL) {
            memcpy(mtdptr, mtds, mno*PKTLAB_MONITOR_DIGEST_LEN);
            mtdptr += mno*PKTLAB_MONITOR_DIGEST_LEN;
            *mtcnt += mno;
        }

        if (priorities != NULL && p != NULL) {
            assert(plen >= 1);
            memcpy(psptr, p, plen);
            psptr[plen-1] = '\0';
            *parrptr = psptr;
            ++parrptr;
            psptr += plen;
            ++*pcnt;
        }
    }

    for (i = 0; i < infonum; ++i)
        free(info_ls[i]);
    free(info_ls);

    return 0;

fail_mtd:
    if (mtdigests != NULL)
        free(*mtdigests);
fail_ftd:
    if (ftdigests != NULL)
        free(*ftdigests);
fail:
    for (i = 0; i < infonum; ++i)
        free(info_ls[i]);
    free(info_ls);

    return -1;
}