// ctrl.c
//

#include <pktlab.h>
#include <pktctrl.h>

#include <ctype.h>
#include <errno.h>
#include <poll.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include "ctrl.h"
#include "private.h"

//
// INTERNAL FUNCTION DECLARATIONS
//

static ssize_t wrapped_read (
	void * fd, void * buf, size_t buflen);
static ssize_t wrapped_write (
	void * fd, const void * buf, size_t nbyte);

//
// EXPORTED FUNCTION DEFINITIONS
//

int pktctrl_raw_session (
	int fd, struct pktctrl_obj * ctrlobj)
{
	trace("%s(state:%d,type:%d,fd:%d)",
		__func__, ctrlobj->state, ctrlobj->type, fd);

	if (fd < 0 || ctrlobj == NULL ||
	    ctrlobj->state != ST_INIT || ctrlobj->type != NEITHER_TYPE)
		return PKTCTRL_INVAL;

	ctrlobj->sockfd = fd;
	ctrlobj->state = ST_ESTABLISHED;
	ctrlobj->type = TCP_TYPE;
	ctrlobj->read_need_write = false; // always false
	ctrlobj->write_need_read = false; // always false

	ctrlobj->reader = pktlab_create_reader(wrapped_read, &ctrlobj->sockfd);
	if (ctrlobj->reader == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->readfunc = wrapped_read;
	ctrlobj->readaux = &ctrlobj->sockfd;

	ctrlobj->writer = pktlab_create_writer(wrapped_write, &ctrlobj->sockfd);
	if (ctrlobj->writer == NULL) {
		ctrlobj->state = ST_ERROR;
		return PKTCTRL_UNKNOWN_FAULT;
	}
	ctrlobj->writefunc = wrapped_write;
	ctrlobj->writeaux = &ctrlobj->sockfd;

	return PKTCTRL_SUCCESS;
}

int pktctrl_prepare_select (
	int * nfds, fd_set * restrict rset, fd_set * restrict wset,
	bool want_read, bool want_write, struct pktctrl_obj * ctrlobj)
{
	trace("%s(state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d,"
		"want_read:%d,want_write:%d)",
		__func__, ctrlobj->state, ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read,
		(int) want_read, (int) want_write);

	if (rset == NULL || wset == NULL || ctrlobj == NULL ||
	    ctrlobj->type == NEITHER_TYPE ||
	    (ctrlobj->state != ST_HANDSHAKING &&
	     ctrlobj->state != ST_ESTABLISHED))
		return PKTCTRL_INVAL;

	if (ctrlobj->state == ST_HANDSHAKING) {
		if (!ctrlobj->hs_need_read && !ctrlobj->hs_need_write) {
			debug("Both hs_need_read and hs_need_write unset, setting both instead");
			ctrlobj->hs_need_read = ctrlobj->hs_need_write = true;
		}

		if (ctrlobj->hs_need_read) {
			FD_SET(ctrlobj->sockfd, rset);
		} else if (ctrlobj->hs_need_write) {
			FD_SET(ctrlobj->sockfd, wset);
		}

		*nfds = PKTLAB_MAX(*nfds, ctrlobj->sockfd+1);
		return PKTCTRL_SUCCESS;
	}

	if (want_read) {
		if (!ctrlobj->read_need_write) {
			FD_SET(ctrlobj->sockfd, rset);
		} else {
			FD_SET(ctrlobj->sockfd, wset);
		}
	}

	if (want_write) {
		if (!ctrlobj->write_need_read) {
			FD_SET(ctrlobj->sockfd, wset);
		} else {
			FD_SET(ctrlobj->sockfd, rset);
		}
	}

	*nfds = PKTLAB_MAX(*nfds, ctrlobj->sockfd+1);
	return PKTCTRL_SUCCESS;
}

int pktctrl_process_select (
	const fd_set * restrict rset, const fd_set * restrict wset,
	struct pktctrl_obj * ctrlobj, bool * readable, bool * writable)
{
	trace("%s(state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d)",
		__func__, ctrlobj->state, ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read);

	if (rset == NULL || wset == NULL || ctrlobj == NULL ||
	    ctrlobj->type == NEITHER_TYPE ||
	    (ctrlobj->state != ST_HANDSHAKING &&
	     ctrlobj->state != ST_ESTABLISHED) ||
	    readable == NULL || writable == NULL)
		return PKTCTRL_INVAL;

	debug("ctrlobj sockfd %d:%s%s", ctrlobj->sockfd,
		FD_ISSET(ctrlobj->sockfd, rset) ? " rset set":"",
		FD_ISSET(ctrlobj->sockfd, wset) ? " wset set":"");

	if (ctrlobj->state == ST_HANDSHAKING) {
		*readable = (bool) FD_ISSET(ctrlobj->sockfd, rset);
		*writable = (bool) FD_ISSET(ctrlobj->sockfd, wset);
		return PKTCTRL_SUCCESS;
	}

	if (ctrlobj->read_need_write) {
		*readable = (bool) FD_ISSET(ctrlobj->sockfd, wset);
	} else {
		*readable = (bool) FD_ISSET(ctrlobj->sockfd, rset);
	}

	if (ctrlobj->write_need_read) {
		*writable = (bool) FD_ISSET(ctrlobj->sockfd, rset);
	} else {
		*writable = (bool) FD_ISSET(ctrlobj->sockfd, wset);
	}

	return PKTCTRL_SUCCESS;
}

int pktctrl_prepare_poll (
	struct pollfd * restrict pfd,
	bool want_read, bool want_write,
	struct pktctrl_obj * ctrlobj)
{
	trace("%s(state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d,"
		"want_read:%d,want_write:%d)",
		__func__, ctrlobj->state, ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read,
		(int) want_read, (int) want_write);

	if (pfd == NULL || ctrlobj == NULL ||
	    ctrlobj->type == NEITHER_TYPE ||
	    (ctrlobj->state != ST_HANDSHAKING &&
	     ctrlobj->state != ST_ESTABLISHED))
		return PKTCTRL_INVAL;

	pfd->fd = ctrlobj->sockfd;
	pfd->events = 0;

	if (ctrlobj->state == ST_HANDSHAKING) {
		if (!ctrlobj->hs_need_read && !ctrlobj->hs_need_write) {
			debug("Both hs_need_read and hs_need_write unset, setting both instead");
			ctrlobj->hs_need_read = ctrlobj->hs_need_write = true;
		}

		if (ctrlobj->hs_need_read) {
			pfd->events |= POLLIN;
		} else if (ctrlobj->hs_need_write) {
			pfd->events |= POLLOUT;
		}

		return PKTCTRL_SUCCESS;
	}

	if (want_read) {
		if (!ctrlobj->read_need_write) {
			pfd->events |= POLLIN;
		} else {
			pfd->events |= POLLOUT;
		}
	}

	if (want_write) {
		if (!ctrlobj->write_need_read) {
			pfd->events |= POLLOUT;
		} else {
			pfd->events |= POLLIN;
		}
	}

	return PKTCTRL_SUCCESS;
}

int pktctrl_process_poll (
	const struct pollfd * restrict pfd,
	struct pktctrl_obj * ctrlobj,
	bool * readable, bool * writable)
{
	trace("%s(pfd:%d,state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d)",
		__func__, pfd->fd, ctrlobj->state, ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read);

	if (pfd == NULL || ctrlobj == NULL ||
	    pfd->fd != ctrlobj->sockfd ||
	    ctrlobj->type == NEITHER_TYPE ||
	    (ctrlobj->state != ST_HANDSHAKING &&
	     ctrlobj->state != ST_ESTABLISHED) ||
	    readable == NULL || writable == NULL)
		return PKTCTRL_INVAL;

	debug("ctrlobj sockfd %d:%s%s", ctrlobj->sockfd,
		(pfd->revents & POLLIN) ? " POLLIN set":"",
		(pfd->revents & POLLOUT) ? " POLLOUT set":"");

	if (ctrlobj->state == ST_HANDSHAKING) {
		*readable = (bool) (pfd->revents & POLLIN);
		*writable = (bool) (pfd->revents & POLLOUT);
		return PKTCTRL_SUCCESS;
	}

	if (ctrlobj->read_need_write) {
		*readable = (bool) (pfd->revents & POLLOUT);
	} else {
		*readable = (bool) (pfd->revents & POLLIN);
	}

	if (ctrlobj->write_need_read) {
		*writable = (bool) (pfd->revents & POLLIN);
	} else {
		*writable = (bool) (pfd->revents & POLLOUT);
	}

	return PKTCTRL_SUCCESS;
}

int pktctrl_read_message (
	struct pktctrl_obj * ctrlobj,
	struct pktlab_message ** msgptr)
{
	trace("%s(state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d,msgptr:%p)",
		__func__, ctrlobj->state,
		ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read,
		msgptr);

	if (ctrlobj == NULL ||
	    ctrlobj->state != ST_ESTABLISHED ||
	    ctrlobj->type == NEITHER_TYPE)
		return -1;

	return pktlab_read_message(ctrlobj->reader, msgptr);
}

int pktctrl_write_message (
	struct pktctrl_obj * ctrlobj,
	const struct pktlab_message * msg)
{
	trace("%s(state:%d,type:%d,sockfd:%d,"
		"hs_need_read:%d,hs_need_write:%d,"
		"read_need_write:%d,write_need_read:%d,msg:%p)",
		__func__, ctrlobj->state,
		ctrlobj->type, ctrlobj->sockfd,
		(int) ctrlobj->hs_need_read, (int) ctrlobj->hs_need_write,
		(int) ctrlobj->read_need_write, (int) ctrlobj->write_need_read, msg);

	if (ctrlobj == NULL ||
	    ctrlobj->state != ST_ESTABLISHED ||
	    ctrlobj->type == NEITHER_TYPE ||
	    msg == NULL)
		return -1;

	return pktlab_write_message(ctrlobj->writer, msg);
}

int pktctrl_flush_write(struct pktctrl_obj * ctrlobj) {
	if (ctrlobj == NULL ||
	    ctrlobj->state != ST_ESTABLISHED ||
	    ctrlobj->type == NEITHER_TYPE)
		return -1;

	return pktlab_flush_writer(ctrlobj->writer);
}

size_t pktctrl_write_unsent(const struct pktctrl_obj * ctrlobj) {
	if (ctrlobj == NULL ||
	    ctrlobj->state != ST_ESTABLISHED ||
	    ctrlobj->type == NEITHER_TYPE)
		return 0;

	return pktlab_writer_unsent(ctrlobj->writer);
}

ssize_t pktctrl_read (
	struct pktctrl_obj * obj,
	void * buf, size_t buflen)
{
	trace("%s(buflen:%zu)", __func__, buflen);

	if (obj == NULL ||
	    obj->state != ST_ESTABLISHED) {
		errno = 0;
		return -1;
	}

	return (obj->readfunc)(obj->readaux, buf, buflen);
}

ssize_t pktctrl_write (
	struct pktctrl_obj * obj,
	const void * buf, size_t nbyte)
{
	trace("%s(nbyte:%zu)", __func__, nbyte);

	if (obj == NULL ||
	    obj->state != ST_ESTABLISHED) {
		errno = 0;
		return -1;
	}

	return (obj->writefunc)(obj->writeaux, buf, nbyte);
}

bool pktctrl_is_established(struct pktctrl_obj * obj) {
	trace("%s(...)", __func__);

	return !(obj == NULL || obj->state != ST_ESTABLISHED);
}

bool pktctrl_is_valid_snikey(const char * snikey) {
	size_t len, i;

	trace("%s(snikey:%s)",
		__func__, (snikey) ? snikey:"(NULL)");

	if (snikey == NULL)
		return false;

	len = strlen(snikey);
	if (len == 0 || len > PKTCTRL_SNIKEY_LEN_MAX) // in [1, 63]
		return false;

	for (i = 0; i < len; ++i)
		// don't allow dashes & dot
		if (!isdigit(snikey[i]) && !islower(snikey[i]))
			return false;

	return true;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

ssize_t wrapped_read (
	void * fd, void * buf, size_t buflen)
{
	trace("%s(sockfd:%d,buflen:%u)",
		__func__, *((int *) fd), (unsigned) buflen);

	if (fd == NULL || buf == NULL)
		return -1;

	return read(*((int *) fd), buf, buflen);
}

ssize_t wrapped_write (
	void * fd, const void * buf, size_t nbyte)
{
	trace("%s(sockfd:%d,nbyte:%u)",
		__func__, *((int *) fd), (unsigned) nbyte);

	if (fd == NULL || buf == NULL || nbyte == 0)
		return -1;

	return write(*((int *) fd), buf, nbyte);
}