// ge_deputy_passive.c
// Passive greet endpoint deputy
// Accept incoming endpoint and create worker threads for exp running.
//

#include <pktlab.h>

#include <assert.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <sys/socket.h>
#include <unistd.h>

#include "net.h"
#include "console.h"
#include "util.h"
#include "worker.h"

#include "ge_deputy.h"

//
// INTERNAL CONSTANTS
//

enum ge_deputy_passive_st {
    NOT_INIT     = 0,
    INIT         = 1,
    G_ONGOING    = 2,
    WRAP_UP      = 3,
    DONE         = 4
};

//
// INTERNAL TYPE DEFINITIONS
//

struct worker_ll_node {
    pthread_t worker_id;
    int worker_fd;
    uint_fast32_t num_id;
    struct worker_ll_node * next;
};

//
// INTERNAL GLOBAL VARIABLES
//

static struct {
    enum ge_deputy_passive_st st;

    char ** exp_prog_n_args;
    struct addrinfo * listen_ai;
    struct pktlab_privatekey * controller_key;
    struct pktctrl_keycert_store * greet_store;

    int endpoint_greet_fd;

    struct {
        uint_fast32_t max_worker_id;
        struct worker_ll_node * worker_ll;
    } worker_info;
} _st;

//
// EXPORTED INTERFACE FUNCTION DECLARATIONS
//

static int ge_deputy_init(
    char ** exp_prog_n_args,
    struct pktlab_privatekey * controller_key,
    struct pktctrl_keycert_store * greet_store,
    void * aux);
static int ge_deputy_start(void);

static int ge_deputy_prepare_select (
    int * nfds, fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until);
static int ge_deputy_process_select (
    const fd_set * rset, const fd_set * wset,
    pktlab_time_t now);

static int ge_deputy_wrapup(void);
static int ge_deputy_cleanup(void);
static bool ge_deputy_is_done(void);

//
// INTERNAL FUNCTION DECLARATIONS
//

static int greet_socket_setup(void);

static int process_new_endpoint(void);
static void free_worker_ll_node (
    struct worker_ll_node * node);
static void process_worker_readable (
    struct worker_ll_node ** pre,
    struct worker_ll_node ** node);

static void greet_endpoint_cleanup(void);

static inline const char * state_str(int st);

//
// EXPORTED FUNCTION DEFINITIONS
//

struct ge_methods get_passive_ge_methods(void) {
    struct ge_methods ret = {
        .init = ge_deputy_init,
        .start = ge_deputy_start,
        .prepare_select = ge_deputy_prepare_select,
        .process_select = ge_deputy_process_select,
        .wrapup = ge_deputy_wrapup,
        .cleanup = ge_deputy_cleanup,
        .is_done = ge_deputy_is_done
    };

    return ret;
}

//
// EXPORTED INTERFACE FUNCTION DEFINITIONS
//

int ge_deputy_init(
    char ** exp_prog_n_args,
    struct pktlab_privatekey * controller_key,
    struct pktctrl_keycert_store * greet_store,
    void * aux)
{
    struct ge_passive_aux * arg = aux; // aux owned by ge_deputy_init

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st != NOT_INIT ||
        exp_prog_n_args == NULL ||
        controller_key == NULL ||
        greet_store == NULL ||
        aux == NULL)
        return -1;

    memset(&_st, 0, sizeof(_st));

    _st.exp_prog_n_args = exp_prog_n_args;
    _st.listen_ai = arg->listen_ai;
    _st.controller_key = controller_key;
    _st.greet_store = greet_store;
    _st.endpoint_greet_fd = -1;
    _st.st = INIT;

    free(aux);
    return 0;
}

int ge_deputy_start(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st != INIT)
        return -1;

    if (greet_socket_setup() < 0) {
        debug("greet_socket_setup failed");
        return -2;
    }

    _st.st = G_ONGOING;
    return 0;
}

int ge_deputy_prepare_select (
    int * nfds, fd_set * rset, fd_set * wset,
    pktlab_time_t * select_until)
{
    struct worker_ll_node * node;

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (nfds == NULL || rset == NULL || wset == NULL ||
        select_until == NULL || _st.st <= INIT)
        return -1;

    switch (_st.st) {
    case G_ONGOING:
        set_fd_set(nfds, rset, _st.endpoint_greet_fd);
        // fall through
    case WRAP_UP:
        for (node = _st.worker_info.worker_ll; node != NULL; node = node->next)
            set_fd_set(nfds, rset, node->worker_fd);

        // don't wait if nothing to wait for
        if (*nfds == 0)
            *select_until = 0;
        break;
    case DONE:
        *select_until = 0; // nothing is going on for us, don't wait
        break;
    default:
        assert(0); // bug
    }

    return 0;
}

int ge_deputy_process_select (
    const fd_set * rset, const fd_set * wset,
    pktlab_time_t now)
{
    struct worker_ll_node * pre;
    struct worker_ll_node * node;
    int rst;

    trace("%s(_st.st:%s,now:%" PRIu64 ")",
        __func__, state_str(_st.st), now);

    if (rset == NULL || wset == NULL ||
        _st.st <= INIT)
        return -1;

    switch (_st.st) {
    case G_ONGOING:
        if (FD_ISSET(_st.endpoint_greet_fd, rset)) {
            debug("endpoint_greet_fd is ready for accept");
            rst = process_new_endpoint();
            if (rst == -11)
                return -2; // should fail immediately
        }
        // fall through
    case WRAP_UP:
        pre = NULL;
        node = _st.worker_info.worker_ll;
        while (node != NULL) {
            if (!FD_ISSET(node->worker_fd, rset)) {
                pre = node;
                node = node->next;
                continue;
            }

            process_worker_readable(&pre, &node);
        }
        // fall through
    case DONE: // do nothing
        break;
    default:
        assert(0); //bug
    }

    // G_ONGOING -> WRAP_UP triggered externally
    // WRAP_UP -> DONE done internally
    if (_st.st == WRAP_UP &&
        _st.worker_info.worker_ll == NULL)
            _st.st = DONE;

    return 0;
}

int ge_deputy_wrapup(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st <= INIT)
        return -1;

    if (_st.st >= WRAP_UP)
        return 0; // already wrapping/wrapped up

    assert(_st.st == G_ONGOING);
    greet_endpoint_cleanup();
    _st.st = WRAP_UP;
    return 0;
}

int ge_deputy_cleanup(void) {
    struct worker_ll_node * node;

    trace("%s(_st.st:%s)", __func__, state_str(_st.st));

    if (_st.st == NOT_INIT ||
        (_st.st >= G_ONGOING &&
         _st.st <= WRAP_UP))
        return -1;

    info("Cleaning up deputy");

    while (_st.worker_info.worker_ll != NULL) {
        node = _st.worker_info.worker_ll->next;
        free_worker_ll_node(_st.worker_info.worker_ll);
        _st.worker_info.worker_ll = node;
    }

    greet_endpoint_cleanup();

    memset(&_st, 0, sizeof(_st));
    _st.st = NOT_INIT;
    return 0;
}

bool ge_deputy_is_done(void) {
    trace("%s(_st.st:%s)", __func__, state_str(_st.st));
    return (_st.st == DONE);
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int greet_socket_setup(void) {
    int rst, ret = -128;

    assert(_st.st == INIT);

    trace("%s()", __func__);

    if (create_welcome_socket(
            _st.listen_ai, &_st.endpoint_greet_fd) < 0) {
        fatal("Failed to create endpoint greeting socket");
        return -1;
    }

    if (set_nonblock(_st.endpoint_greet_fd, true) < 0) {
        fatal("Failed to set endpoint greeting socket non-block");
        ret = -2;
        goto fail_sock;
    }

    rst = set_close_on_exec(_st.endpoint_greet_fd);
    if (rst < 0) {
        fatal("Failed to set endpoint greeting socket close-on-exec");
        ret = -3;
        goto fail_sock;
    }

    return 0;

fail_sock:
    close(_st.endpoint_greet_fd);
    _st.endpoint_greet_fd = -1;
    return ret;
}

int process_new_endpoint(void) {
    int endpoint_fd, rst, err, ret = -128, fds[2];
    struct sockaddr_storage sa;
    struct worker_arg * wa;
    sigset_t oset;
    pthread_t worker_id;
    struct worker_ll_node * node;
    uint_fast32_t worker_num_id;

    assert(_st.st == G_ONGOING);

    trace("%s()", __func__);

    if (accept_new_socket(
            _st.endpoint_greet_fd,
            &endpoint_fd, &sa, &err) < 0) {
        warn("Failed to accept new endpoint (%s), ignoring",
            strerror(err));
        return -1;
    }

    if (print_info_addr("New endpoint from ", &sa) < 0) {
        warn("Failed to print new endpoint address, "
            "skipping this endpoint");
        ret = -2;
        goto fail_accept;
    }

    if (set_close_on_exec(endpoint_fd) < 0) {
        warn("Failed to set new endpoint socket close-on-exec, "
            "skipping this endpoint");
        ret = -3;
        goto fail_accept;
    }

    if (set_nonblock(endpoint_fd, true) < 0) {
        warn("Failed to set new endpoint socket non-blocking, "
            "skipping this endpoint");
        ret = -4;
        goto fail_accept;
    }

    if (set_keepalive(endpoint_fd) < 0) {
        warn("Failed to set new endpoint socket keepalive, "
            "skipping this endpoint");
        ret = -5;
        goto fail_accept;
    }

    // spawn worker thread
    if (get_stream_socketpair(fds, &err) < 0) {
        warn("Failed to create manager-worker socket pair (%s), "
            "skipping this endpoint", strerror(err));
        ret = -6;
        goto fail_accept;
    }

    if (set_close_on_exec(fds[0]) < 0 ||
        set_close_on_exec(fds[1]) < 0) {
        warn("Failed to set internal socket pair close-on-exec (%s), "
            "skipping this endpoint", strerror(errno));
        ret = -7;
        goto fail_fds;
    }

    if (set_nonblock(fds[0], true) < 0 ||
        set_nonblock(fds[1], true) < 0) {
        warn("Failed to set internal socket pair non-blocking (%s), "
            "skipping this endpoint", strerror(errno));
        ret = -8;
        goto fail_fds;
    }

    worker_num_id = _st.worker_info.max_worker_id+1;
    _st.worker_info.max_worker_id = worker_num_id;
    // update _st here, so if fail afterwards id is not retaken

    wa = safe_malloc(sizeof(*wa));
    wa->endpoint_fd = endpoint_fd;
    wa->manager_fd = fds[1]; // fds[1] is for worker
    wa->worker_id = worker_num_id;
    wa->exp_prog_n_args = _st.exp_prog_n_args;
    wa->k = _st.controller_key;
    wa->store = _st.greet_store;

    if (set_sigmask_block_sigint(&oset) < 0) {
        warn("Failed to block SIGINT, skipping this endpoint");
        ret = -9;
        goto fail_wa;
    }

    rst = pthread_create(
        &worker_id, NULL, worker_run, wa);
    if (rst != 0) {
        warn("Failed to create worker thread (%s), "
            "skipping this endpoint", strerror(rst));
        ret = -10;
        goto fail_wa;
    }

    if (restore_sigmask(&oset) < 0) {
        fatal("Failed to restore sigmask, aborting");
        ret = -11;
        goto fail_post_create;
    }

    rst = pthread_detach(worker_id); // no joining afterwards
    if (rst != 0) {
        warn("Failed to detach worker thread (%s), "
            "early worker termination", strerror(rst));
        ret = -12;
        // we only "signal" thread should die here
        goto fail_post_create;
    }

    node = safe_malloc(sizeof(*node));
    node->worker_id = worker_id;
    node->worker_fd = fds[0]; // fds[0] is for manager
    node->num_id = worker_num_id;
    node->next = _st.worker_info.worker_ll;
    _st.worker_info.worker_ll = node;

    info("Assigning endpoint to worker %" PRIuFAST32,
        worker_num_id);
    return 0;

// for failing until pthread_create
// as deputy still owns everything

fail_wa:
    free(wa);
fail_fds:
    close(fds[0]);
    close(fds[1]);
fail_accept:
    close(endpoint_fd);
    return ret;

// for failing after pthread_create
// only free the part deputy owns.
// the others are freed by worker

fail_post_create:
    close(fds[0]);
    return ret;
}

void free_worker_ll_node (
    struct worker_ll_node * node)
{
    assert(node != NULL);

    trace("%s()", __func__);

    close(node->worker_fd);
    free(node);
    return;
}

void process_worker_readable (
    struct worker_ll_node ** pre,
    struct worker_ll_node ** node)
{
    int rst;
    uint8_t msg;
    struct worker_ll_node * next;

    assert(_st.st == G_ONGOING || _st.st == WRAP_UP);
    assert(pre != NULL && node != NULL && *node != NULL);

    trace("%s(...)", __func__);

    rst = recv_worker_msg((*node)->worker_fd, &msg);
    switch (rst) {
    case 1: // received msg or EOF
        if (msg == WORKER_TLS_COMPLETED) {
            info("Worker %" PRIuFAST32 " TLS handshake completed",
                (*node)->num_id);
            *pre = *node;
            *node = (*node)->next;
            return; // worker notification only, carry on
        }

        assert(msg == WORKER_EOF);
        info("Worker %" PRIuFAST32 " socket EOF, "
            "cleaning up worker", (*node)->num_id);
        break; // free worker node
    case 0: // try again
        *pre = *node; // for next node
        *node = (*node)->next;
        return;
    default:
        assert(rst < 0); // error
        warn("Unexpected error (%s) when reading worker (%" PRIuFAST32
            ") socket, cleaning up worker",
            strerror(errno), (*node)->num_id);
        break; // free worker node
    }

    if (*pre == NULL) {
        next = _st.worker_info.worker_ll = (*node)->next;
    } else {
        next = (*pre)->next = (*node)->next;
    }

    free_worker_ll_node(*node);
    *node = next;
    return; // pre stays the same
}

void greet_endpoint_cleanup(void) {
    if (_st.st == NOT_INIT ||
        _st.endpoint_greet_fd < 0)
        return;

    close(_st.endpoint_greet_fd);

    _st.endpoint_greet_fd = -1;
}

const char * state_str(int st) {
    static const char * sstr[] = {
        [NOT_INIT]     = "NOT_INIT",
        [INIT]         = "INIT",
        [G_ONGOING]    = "G_ONGOING",
        [WRAP_UP]      = "WRAP_UP",
        [DONE]         = "DONE"
    };

    assert(st >= 0 && st < sizeof(sstr)/sizeof(*sstr));
    return safe_str(sstr[st]);
}