// cert.c
//

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <pktlab.h>
#include <pktconf.h>

#include "cert.h"
#include "console.h"
#include "util.h"

//
// EXPORTED FUNCTION DEFINITIONS
//

int ll_load_cert_chain (
    const struct pktconf_sep_str_ll * sep_list,
    struct pktlab_certificate *** certs,
    uint_fast32_t * certno)
{
    struct pktconf_sep_str_entry * node;
    uint_fast32_t i;
    size_t totalmem;
    int ret = -128;

    assert(certs != NULL && certno != NULL);

    if (sep_list == NULL) {
        *certs = NULL;
        *certno = 0;
        return 0;
    }

    totalmem = sizeof(**certs)*sep_list->len;
    if (totalmem / sizeof(**certs) != sep_list->len)
        return -1; // overflow

    *certs = safe_malloc(totalmem);
    *certno = 0;
    node = sep_list->head;

    for (i = 0; i < sep_list->len; ++i) {
        assert(node->entry != NULL);

        if (node->entrylen == 0) {
            debug("Bad cert chain ll entry content");
            ret = -2;
            goto fail;
        }

        if (load_cert(node->entry, &(*certs)[i]) < 0) {
            debug("Failed to load from %s", node->entry);
            ret = -3;
            goto fail;
        }

        ++*certno;
        node = node->next;
    }

    return 0;

fail:
    free_certs(*certs, *certno);
    return ret;
}

void free_certs (
    struct pktlab_certificate ** certs,
    uint_fast32_t certno)
{
    uint_fast32_t i;

    assert(certs != NULL);

    for (i = 0; i < certno; ++i)
        pktlab_cleanup_cert(certs[i]);

    free(certs);
}

int reorder_certs (
    struct pktlab_privatekey * controller_key,
    struct pktlab_certificate ** certs,
    uint_fast32_t certno)
{
    uint_fast32_t i;
    struct pktlab_publickey * pubkey;
    uint8_t sid[PKTLAB_KEYID_LEN];
    struct pktlab_cert_info * info;
    struct pktlab_certificate * cert;
    int rst, ret = -128;

    assert(
        controller_key != NULL &&
        certs != NULL && certno != 0);

    pubkey = pktlab_get_publickey_from_privatekey(
        controller_key);
    if (pubkey == NULL) {
        debug("Cannot get public key for controller private key");
        return -1;
    }

    if (pktlab_get_key_identifier(
            pubkey, sid, sizeof(sid)) < 0) {
        debug("Cannot get key identifier for controller public key");
        ret = -2;
        goto fail;
    }

    for (i = 0; i < certno; ++i) {
        rst = pktlab_decode_certificate(
            certs[i], NULL, &info);
        if (rst != PKTLAB_DECODE_SUCCESS) {
            debug("Cert decoding failed: %d", rst);
            ret = -3;
            goto fail;
        }

        if (memcmp(info->subject_key_id, sid, sizeof(sid)) == 0) {
            free(info);
            break;
        }

        free(info);
    }

    if (i == certno) {
        debug("No cert with matching sid found");
        ret = -4;
        goto fail;
    }

    // swapping of matching cert to first in certs
    if (i != 0) {
        cert = certs[i];
        certs[i] = certs[0];
        certs[0] = cert;
    }

    ret = 0;

fail:
    pktlab_cleanup_publickey(pubkey);
    return ret;
}

int load_cert (
    const char * path,
    struct pktlab_certificate ** cert)
{
    char * rawcert;
    size_t rawcertlen;
    int ret = -128;

    assert(path != NULL && cert != NULL);

    trace("%s(path:%s)", __func__, path);

    if (read_whole_file(
            path, (uint8_t **) &rawcert, &rawcertlen) < 0) {
        debug("Cannot read whole file content for %s", path);
        ret = -1;
        goto fail;
    }

    if (rawcertlen > UINT_FAST32_MAX) {
        debug("Overflow");
        ret = -2;
        goto fail_raw;
    }

    *cert = pktlab_create_cert();

    if (pktlab_load_cert_pem(
            rawcert, rawcertlen, *cert) < 0) {
        debug("Failed to load certificate");
        ret = -3;
        goto fail_cert;
    }

    free(rawcert);
    return 0;

fail_cert:
    pktlab_cleanup_cert(*cert);
fail_raw:
    free(rawcert);
fail:
    return ret;
}