// config.c
//

#include "config.h"

#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>

#include "console.h"
#include "util.h"
#include "net.h"

#define BUF_SIZ 1024

//
// Guide on adding new fields
// (1) Add enum entry in "enum me_init_keys" for read_config_line parsing
// identification (2) Add corresponding entry in _key_dict, specifying
//     (a) keyname (i.e. the option name in config file), and
//     (b) identifier (i.e. the new entry in "enum me_init_keys")
// (3) Add corresponding fields in struct config of config.h
// (4) Change config_initialize function for added struct config fields initing
// (5) Change read_config_line function for config line parsing and
//     struct config fields updating w.r.t. the new keyname of _key_dict.
//     Remember to add the new enum me_init_keys entry in assert.
// (6) Add debug msg in config_read function for printing config content
//

//
// INTERNAL TYPE DEFINITIONS
//

enum me_init_keys {
    NETWORK_BUFMAX,
    TRANSPROTO_SUP,
    VMEM_LADDR_STR_V4,
    VMEM_LADDR_STR_V6,
    VMEM_LDNSADDR_STR_V4,
    VMEM_LDNSADDR_STR_V6,
    HOST_ID,
    BROKER_ADDR,
    BROKER_CHAN,
    AUTO_EXTRACT_CHAN,
    ENDPOINT_CERT_PATH,
    BROKER_CHAIN_ROOT_PATH,
    EOP_PUBK_PATH,
    SUBPRIV_CHAIN_PATH,
    CONTROLLER_CHAIN_PATH,
    ENDPOINT_PRIVK_PATH,
    NO_SUB_MODE,
    UKN
};

typedef int (*value_parser_t)(const char *value, uint_fast32_t valuelen,
                              uint_fast32_t lineno);

struct key_dict_entry {
    const char *const keyname;
    uint_fast32_t keylen;
    const enum me_init_keys key;
};

//
// INTERNAL GLOBAL VARIABLES
//

#define NETWORK_BUFMAX_STR "BufMax"
#define TRANSPROTO_SUP_STR "TransprotoSup"
#define VMEM_LADDR_STR_V4_STR "VMemLocalAddrStrIPv4"
#define VMEM_LADDR_STR_V6_STR "VMemLocalAddrStrIPv6"
#define VMEM_LDNSADDR_STR_V4_STR "VMemLocalDNSAddrStrIPv4"
#define VMEM_LDNSADDR_STR_V6_STR "VMemLocalDNSAddrStrIPv6"
#define HOST_ID_STR "HostID"
#define BROKER_ADDR_STR "BrokerAddr"
#define BROKER_CHAN_STR "BrokerChannel"
#define AUTO_EXTRACT_CHAN_STR "AutoExtractChannel"
#define ENDPOINT_AGENT_CERT_PATH_STR "EndpointAgentCertPath"
#define BROKER_CHAIN_ROOT_PATH_STR "BrokerChainRootPath"
#define EOP_PUBK_PATH_STR "EndpointOpPubKeyPath"
#define SUBPRIV_CHAIN_PATH_STR "SubPrivCertChainPath"
#define CONTROLLER_CHAIN_PATH_STR "ControllerCertChainPath"
#define ENDPOINT_PRIVK_PATH_STR "EndpointPrivKey"
#define NO_SUB_MODE_STR "NoSubMode"

static const struct key_dict_entry _key_dict[] = {
    {NETWORK_BUFMAX_STR, sizeof(NETWORK_BUFMAX_STR) - 1, NETWORK_BUFMAX},
    {TRANSPROTO_SUP_STR, sizeof(TRANSPROTO_SUP_STR) - 1, TRANSPROTO_SUP},
    {VMEM_LADDR_STR_V4_STR, sizeof(VMEM_LADDR_STR_V4_STR) - 1,
     VMEM_LADDR_STR_V4},
    {VMEM_LADDR_STR_V6_STR, sizeof(VMEM_LADDR_STR_V6_STR) - 1,
     VMEM_LADDR_STR_V6},
    {VMEM_LDNSADDR_STR_V4_STR, sizeof(VMEM_LDNSADDR_STR_V4_STR) - 1,
     VMEM_LDNSADDR_STR_V4},
    {VMEM_LDNSADDR_STR_V6_STR, sizeof(VMEM_LDNSADDR_STR_V6_STR) - 1,
     VMEM_LDNSADDR_STR_V6},
    {HOST_ID_STR, sizeof(HOST_ID_STR) - 1, HOST_ID},
    {BROKER_ADDR_STR, sizeof(BROKER_ADDR_STR) - 1, BROKER_ADDR},
    {BROKER_CHAN_STR, sizeof(BROKER_CHAN_STR) - 1, BROKER_CHAN},
    {AUTO_EXTRACT_CHAN_STR, sizeof(AUTO_EXTRACT_CHAN_STR) - 1,
     AUTO_EXTRACT_CHAN},
    {ENDPOINT_AGENT_CERT_PATH_STR, sizeof(ENDPOINT_AGENT_CERT_PATH_STR) - 1,
     ENDPOINT_CERT_PATH},
    {BROKER_CHAIN_ROOT_PATH_STR, sizeof(BROKER_CHAIN_ROOT_PATH_STR) - 1,
     BROKER_CHAIN_ROOT_PATH},
    {EOP_PUBK_PATH_STR, sizeof(EOP_PUBK_PATH_STR) - 1, EOP_PUBK_PATH},
    {SUBPRIV_CHAIN_PATH_STR, sizeof(SUBPRIV_CHAIN_PATH_STR) - 1,
     SUBPRIV_CHAIN_PATH},
    {CONTROLLER_CHAIN_PATH_STR, sizeof(CONTROLLER_CHAIN_PATH_STR) - 1,
     CONTROLLER_CHAIN_PATH},
    {ENDPOINT_PRIVK_PATH_STR, sizeof(ENDPOINT_PRIVK_PATH_STR) - 1,
     ENDPOINT_PRIVK_PATH},
    {NO_SUB_MODE_STR, sizeof(NO_SUB_MODE_STR) - 1,
     NO_SUB_MODE},
    {NULL, 0, UKN}

};

//
// EXPORTED GLOBAL VARIABLES
//

struct config _config;

//
// INTERNAL FUNCTION DECLARATIONS
//

static enum me_init_keys check_dict(const char *keyname, uint_fast32_t keylen);
static int load_config(const char *key, uint_fast32_t keylen, const char *value,
                       uint_fast32_t valuelen, uint_fast32_t lineno);

static int parse_bool(const char *value, uint_fast32_t valuelen,
                      bool *field_set, bool *field);
static int parse_addr(const char *value, uint_fast32_t valuelen,
                      const char **field);
static int parse_filepath(const char *value, uint_fast32_t valuelen,
                          const char **field);
static int parse_filepath_ls(const char *value, uint_fast32_t valuelen,
                             const struct pktconf_sep_str_ll **field);
static int parse_ull(const char *value, uint_fast32_t valuelen,
                     unsigned long long *field);
static int parse_hex_ls(const char *value, uint_fast32_t valuelen,
                        const struct pktconf_sep_str_ll **field);

static int parse_network_bufmax(const char *value, uint_fast32_t valuelen,
                                uint_fast32_t lineno);
static int parse_transproto_sup(const char *value, uint_fast32_t valuelen,
                                uint_fast32_t lineno);
static int parse_vmem_laddr_str_v4(const char *value, uint_fast32_t valuelen,
                                   uint_fast32_t lineno);
static int parse_vmem_laddr_str_v6(const char *value, uint_fast32_t valuelen,
                                   uint_fast32_t lineno);
static int parse_vmem_ldnsaddr_str_v4(const char *value, uint_fast32_t valuelen,
                                      uint_fast32_t lineno);
static int parse_vmem_ldnsaddr_str_v6(const char *value, uint_fast32_t valuelen,
                                      uint_fast32_t lineno);
static int parse_host_id(const char *value, uint_fast32_t valuelen,
                         uint_fast32_t lineno);
static int parse_broker_addr(const char *value, uint_fast32_t valuelen,
                             uint_fast32_t lineno);
static int parse_broker_chan(const char *value, uint_fast32_t valuelen,
                             uint_fast32_t lineno);
static int parse_auto_extract_chan(const char *value, uint_fast32_t valuelen,
                                   uint_fast32_t lineno);
static int parse_endpoint_agent_cert_path(const char *value,
                                          uint_fast32_t valuelen,
                                          uint_fast32_t lineno);
static int parse_broker_chain_root_path(const char *value,
                                        uint_fast32_t valuelen,
                                        uint_fast32_t lineno);
static int parse_eop_pubk_path(const char *value, uint_fast32_t valuelen,
                               uint_fast32_t lineno);
static int parse_subpriv_chain_path(const char *value, uint_fast32_t valuelen,
                                    uint_fast32_t lineno);
static int parse_controller_chain_path(const char *value,
                                       uint_fast32_t valuelen,
                                       uint_fast32_t lineno);
static int parse_endpoint_privk_path(const char *value, uint_fast32_t valuelen,
                                     uint_fast32_t lineno);
static int parse_no_sub_mode(const char *value, uint_fast32_t valuelen,
                             uint_fast32_t lineno);

//
// EXPORTED FUNCTION DEFINITIONS
//

int config_initialize(void) {
    trace("%s()", __func__);

    memset(&_config, 0, sizeof(_config));
    _config.network_bufmax = -1;
    _config.raw_sup = false;
    _config.tcp_sup = false;
    _config.udp_sup = false;
    _config.auto_extract_channel = false;
    _config.no_sub_mode = false;
    _config.state = CONFIG_INIT;
    return 0;
}

int config_read(const char *path, uint_fast32_t *errlineno) {
    struct pktconf_option_ll *ll;
    struct pktconf_opt_ll_node *node;
    uint_fast32_t i;

    trace("%s(path:%s)", __func__, safe_str(path));

    if (_config.state != CONFIG_INIT) {
        debug("Bad config state");
        return -1; // only read when in CONFIG_INIT state
    }

    if (path == NULL) {
        debug("No config path provided");
        return -2;
    }

    if (errlineno != NULL)
        *errlineno = 0; // 0 implies no line read

    if (pktconf_load_option(path, &ll) < 0) {
        debug("pktconf_load_option failed");
        return -3;
    }

    node = ll->head;
    node = ll->head;
    for (i = 0; i < ll->len; ++i) {
        if (load_config(node->key, node->keylen, node->value, node->valuelen,
                        node->lineno) < 0) {
            debug("load_config(key:%s,value:%s) failed at line %u", node->key,
                  node->value, (unsigned)node->lineno);
            *errlineno = node->lineno;
            pktconf_cleanup_option_ll(ll);
            config_cleanup();
            return -4;
        }

        node = node->next;
    }

    _config.state = CONFIG_READY;

    debug("config.BufMax:\t%lld", _config.network_bufmax);
    debug("config.TransprotoSup: raw(%s), tcp(%s), udp(%s)",
          (_config.raw_sup) ? "T" : "F", (_config.tcp_sup) ? "T" : "F",
          (_config.udp_sup) ? "T" : "F");
    debug("config.VMemLocalAddrStrIPv4:\t%s",
          safe_str(_config.vmem_laddr_str_v4));
    debug("config.VMemLocalAddrStrIPv6:\t%s",
          safe_str(_config.vmem_laddr_str_v6));
    debug("config.VMemLocalDNSAddrStrIPv4:\t%s",
          safe_str(_config.vmem_ldnsaddr_str_v4));
    debug("config.VMemLocalDNSAddrStrIPv6:\t%s",
          safe_str(_config.vmem_ldnsaddr_str_v6));
    debug("config.HostID:\t%s", safe_str(_config.host_id));
    debug("config.%s:\t%s", BROKER_ADDR_STR, safe_str(_config.broker_addr));
    debug("config.%s:\t%s", ENDPOINT_AGENT_CERT_PATH_STR,
          safe_str(_config.endpoint_agent_cert_path));
    debug("config.%s:\t%s", BROKER_CHAIN_ROOT_PATH_STR,
          str_from_pktconf_ll(_config.broker_chain_troot_path_ll, ','));
    debug("config.%s:\t%s", BROKER_CHAN_STR,
          str_from_pktconf_ll(_config.broker_channel_ll, ','));
    debug("config.%s:\t%d", AUTO_EXTRACT_CHAN_STR,
          (int) _config.auto_extract_channel);
    debug("config.%s:\t%s", SUBPRIV_CHAIN_PATH_STR,
          str_from_pktconf_ll(_config.subpriv_chain_path_ll, ','));
    debug("config.%s:\t%s", CONTROLLER_CHAIN_PATH_STR,
          str_from_pktconf_ll(_config.controller_chain_path_ll, ','));
    debug("config.%s:\t%s", EOP_PUBK_PATH_STR,
          str_from_pktconf_ll(_config.eop_pubk_path_ll, ','));
    debug("config.%s:\t%s", ENDPOINT_PRIVK_PATH_STR,
          safe_str(_config.endpoint_privk_path));
    debug("config.%s:\t%d", NO_SUB_MODE_STR,
          (int) _config.no_sub_mode);
    return 0;
}

bool is_config_presence_correct(int *err) {
    assert(err != NULL);
    *err = 0;

    if (_config.state != CONFIG_READY) {
        *err = -1;
        return false;
    }

    if (_config.vmem_laddr_str_v4 == NULL &&
        _config.vmem_laddr_str_v6 == NULL) {
        *err = -2;
        return false;
    }

    if ((_config.raw_sup || _config.udp_sup || _config.tcp_sup) == false) {
        *err = -3;
        return false;
    }

    if (_config.eop_pubk_path_ll == NULL && _config.endpoint_privk_path == NULL) {
        *err = -4;
        return false;
    }

    if (!_config.no_sub_mode &&
        (_config.broker_addr == NULL ||
         _config.endpoint_agent_cert_path == NULL ||
         _config.broker_chain_troot_path_ll == NULL)) {
        *err = -5;
        return false;
    }

    if (_config.controller_chain_path_ll == NULL) {
        *err = -6;
        return false;
    }
    return true;
}

void config_cleanup(void) {
    trace("%s()", __func__);

    if (_config.state == CONFIG_NOT_INIT)
        return;

    free((void *)_config.vmem_laddr_str_v4);
    free((void *)_config.vmem_laddr_str_v6);
    free((void *)_config.vmem_ldnsaddr_str_v4);
    free((void *)_config.vmem_ldnsaddr_str_v6);
    free((void *)_config.host_id);
    free((void *)_config.broker_addr);
    free((void *)_config.endpoint_agent_cert_path);
    free((void *)_config.endpoint_privk_path);

    pktconf_cleanup_sep_str_ll((void *)_config.broker_chain_troot_path_ll);
    pktconf_cleanup_sep_str_ll((void *)_config.broker_channel_ll);
    pktconf_cleanup_sep_str_ll((void *)_config.subpriv_chain_path_ll);
    pktconf_cleanup_sep_str_ll((void *)_config.eop_pubk_path_ll);

    _config.state = CONFIG_NOT_INIT;
    return;
}

//
// INTERNAL FUNCTION DEFINITIONS
//

enum me_init_keys check_dict(const char *keyname, uint_fast32_t keylen) {
    uint_fast32_t i;

    assert(keyname != NULL);

    for (i = 0; _key_dict[i].keyname != NULL; ++i)
        if (keylen == _key_dict[i].keylen &&
            strncmp(keyname, _key_dict[i].keyname, keylen) == 0)
            return _key_dict[i].key;

    return UKN;
}

int load_config(const char *key, uint_fast32_t keylen, const char *value,
                uint_fast32_t valuelen, uint_fast32_t lineno) {
    enum me_init_keys k;
    static const value_parser_t parsers[] = {
        [NETWORK_BUFMAX] = &parse_network_bufmax,
        [TRANSPROTO_SUP] = &parse_transproto_sup,
        [VMEM_LADDR_STR_V4] = &parse_vmem_laddr_str_v4,
        [VMEM_LADDR_STR_V6] = &parse_vmem_laddr_str_v6,
        [VMEM_LDNSADDR_STR_V4] = &parse_vmem_ldnsaddr_str_v4,
        [VMEM_LDNSADDR_STR_V6] = &parse_vmem_ldnsaddr_str_v6,
        [HOST_ID] = &parse_host_id,
        [BROKER_ADDR] = &parse_broker_addr,
        [BROKER_CHAN] = &parse_broker_chan,
        [AUTO_EXTRACT_CHAN] = &parse_auto_extract_chan,
        [ENDPOINT_CERT_PATH] = &parse_endpoint_agent_cert_path,
        [BROKER_CHAIN_ROOT_PATH] = &parse_broker_chain_root_path,
        [EOP_PUBK_PATH] = &parse_eop_pubk_path,
        [SUBPRIV_CHAIN_PATH] = &parse_subpriv_chain_path,
        [CONTROLLER_CHAIN_PATH] = &parse_controller_chain_path,
        [ENDPOINT_PRIVK_PATH] = &parse_endpoint_privk_path,
        [NO_SUB_MODE] = &parse_no_sub_mode,
        [UKN] = NULL,
    };
    value_parser_t parser;

    assert(key != NULL && value != NULL);

    trace("%s(key:%s,keylen:%" PRIuFAST32 ",value:%s,valuelen:%" PRIuFAST32 ")",
          __func__, key, keylen, value, valuelen);

    k = check_dict(key, keylen);

    assert(k >= 0 && k < sizeof(parsers) / sizeof(*parsers));

    parser = parsers[k];

    if (parser != NULL) {
        if (parser(value, valuelen, lineno) != 0) {
            fatal("Failed to parse field in config file "
                  "(key:'%s',value:'%s') at line %" PRIuFAST32 ", aborting",
                  key, value, lineno);
            return -1;
        } else
            return 0;
    } else {
        warn("Unknown field in config file "
             "(key:'%s',value:'%s') at line %" PRIuFAST32 ", ignoring",
             key, value, lineno);
        return 0;
    }
}

int parse_bool(const char *value, uint_fast32_t valuelen, bool *field_set,
               bool *field) {
    assert(value != NULL && field_set != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (*field_set)
        return 0; // use last entry

    if (valuelen != 1)
        return -1;

    switch (*value) {
    case '0':
        *field_set = true;
        *field = false;
        break;
    case '1':
        *field_set = true;
        *field = true;
        break;
    default:
        return -1;
    }

    return 0;
}

int parse_addr(const char *value, uint_fast32_t valuelen, const char **field) {
    char *addr;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no address specified, remain NULL

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(value, valuelen, &addr) < 0) {
        debug("copy_str failed");
        return -1;
    }

    if (parse_n_getaddrinfo(addr, 0, NULL) < 0) {
        debug("parse_n_getaddrinfo failed, possibly bad address");
        free(addr);
        return -1;
    }

    *field = addr;
    return 0;
}

int parse_filepath(const char *value, uint_fast32_t valuelen,
                   const char **field) {
    char *filepath;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(value, valuelen, &filepath) < 0) {
        debug("copy_str failed");
        return -1;
    }

    if (!is_valid_filepath(filepath)) {
        debug("Bad filepath: %s", filepath);
        free(filepath);
        return -1;
    }

    *field = filepath;
    return 0;
}

int parse_filepath_ls(const char *value, uint_fast32_t valuelen,
                      const struct pktconf_sep_str_ll **field) {
    struct pktconf_sep_str_ll *ll;
    struct pktconf_sep_str_entry *entry;
    uint_fast32_t i;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (pktconf_parse_sep_str(value, valuelen, ',', UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (!is_valid_filepath(entry->entry)) {
            debug("Bad filepath: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    *field = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_ull(const char *value, uint_fast32_t valuelen,
              unsigned long long *field) {
    unsigned long long u;
    char *endptr;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (valuelen == 0)
        return 0; // nothing specified, remain NULL

    errno = 0;
    u = strtoull(value, &endptr, 10);
    if (errno != 0 || endptr != value + valuelen) {
        debug("Bad ull: %s", value);
        return -1; // conversion failed/out-of-range
    }

    *field = u;
    return 0;
}

int parse_hex_ls(const char *value, uint_fast32_t valuelen,
                 const struct pktconf_sep_str_ll **field) {
    struct pktconf_sep_str_ll *ll;
    struct pktconf_sep_str_entry *entry;
    uint_fast32_t i;

    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);

    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0; // no path specified, remain NULL

    if (pktconf_parse_sep_str(value, valuelen, ',', UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        if (entry->entrylen > 0 && !is_valid_hexstr(entry->entry)) {
            debug("Bad hex str: %s", entry->entry);
            goto fail;
        }

        entry = entry->next;
    }

    *field = ll;
    return 0;
fail:
    pktconf_cleanup_sep_str_ll(ll);
    return -1;
}

int parse_str(const char *value, uint_fast32_t valuelen, const char **field) {
    char *str;
    assert(value != NULL && field != NULL);

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ")", __func__, value, valuelen);
    if (*field != NULL)
        return 0; // use the last non-empty entry if multiple exist

    if (valuelen == 0)
        return 0;

    if (valuelen >= SIZE_MAX)
        return -1; // overflow

    if (copy_str(value, valuelen, &str) < 0) {
        debug("copy_str failed");
        return -1;
    }

    *field = str;
    return 0;
}

int parse_network_bufmax(const char *value, uint_fast32_t valuelen,
                         uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);

    return parse_ull(value, valuelen, &_config.network_bufmax);
}

int parse_transproto_sup(const char *value, uint_fast32_t valuelen,
                         uint_fast32_t lineno) {
    struct pktconf_sep_str_ll *ll;
    uint_fast32_t i;
    struct pktconf_sep_str_entry *entry;
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);

    if (pktconf_parse_sep_str(value, valuelen, ',', UINT_FAST32_MAX, &ll) < 0) {
        debug("pktconf_parse_sep_str failed");
        return -1;
    }

    entry = ll->head;
    for (i = 0; i < ll->len; ++i) {
        assert(entry != NULL);
        // FIXME: 3 or 4
        if (entry->entrylen == 3) {
            _config.tcp_sup =
                _config.tcp_sup || !strncasecmp(entry->entry, "TCP", 3);
            _config.udp_sup =
                _config.udp_sup || !strncasecmp(entry->entry, "UDP", 3);
            _config.raw_sup =
                _config.raw_sup || !strncasecmp(entry->entry, "RAW", 3);
        } else {
            debug("unknown proto %s", entry->entry);
        }
        entry = entry->next;
    }
    pktconf_cleanup_sep_str_ll(ll);
    return 0;
}

int parse_vmem_laddr_str_v4(const char *value, uint_fast32_t valuelen,
                            uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_str(value, valuelen, &_config.vmem_laddr_str_v4);
}

int parse_vmem_laddr_str_v6(const char *value, uint_fast32_t valuelen,
                            uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_str(value, valuelen, &_config.vmem_laddr_str_v6);
}

int parse_vmem_ldnsaddr_str_v4(const char *value, uint_fast32_t valuelen,
                               uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_str(value, valuelen, &_config.vmem_ldnsaddr_str_v4);
}

int parse_vmem_ldnsaddr_str_v6(const char *value, uint_fast32_t valuelen,
                               uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_str(value, valuelen, &_config.vmem_ldnsaddr_str_v6);
}

int parse_host_id(const char *value, uint_fast32_t valuelen,
                  uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_str(value, valuelen, &_config.host_id);
}

int parse_broker_addr(const char *value, uint_fast32_t valuelen,
                      uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_addr(value, valuelen, &_config.broker_addr);
}

int parse_broker_chan(const char *value, uint_fast32_t valuelen,
                      uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_hex_ls(value, valuelen, &_config.broker_channel_ll);
}

int parse_auto_extract_chan(const char *value, uint_fast32_t valuelen,
                            uint_fast32_t lineno) {
    bool is_set = false; // undefined behavior if multiple such option specified

    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_bool(value, valuelen, &is_set, &_config.auto_extract_channel);
}

int parse_endpoint_agent_cert_path(const char *value, uint_fast32_t valuelen,
                                   uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath(value, valuelen, &_config.endpoint_agent_cert_path);
}

int parse_broker_chain_root_path(const char *value, uint_fast32_t valuelen,
                                 uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath_ls(value, valuelen,
                             &_config.broker_chain_troot_path_ll);
}

int parse_eop_pubk_path(const char *value, uint_fast32_t valuelen,
                        uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath_ls(value, valuelen, &_config.eop_pubk_path_ll);
}

int parse_subpriv_chain_path(const char *value, uint_fast32_t valuelen,
                             uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath_ls(value, valuelen, &_config.subpriv_chain_path_ll);
}

int parse_controller_chain_path(const char *value, uint_fast32_t valuelen,
                                uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath_ls(value, valuelen,
                             &_config.controller_chain_path_ll);
}

int parse_endpoint_privk_path(const char *value, uint_fast32_t valuelen,
                              uint_fast32_t lineno) {
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_filepath(value, valuelen, &_config.endpoint_privk_path);
}

int parse_no_sub_mode(const char *value, uint_fast32_t valuelen,
                      uint_fast32_t lineno) {
    bool field_set = false;
    trace("%s(value:%s,valuelen:%" PRIuFAST32 ",lineno:%" PRIuFAST32 ")",
          __func__, value, valuelen, lineno);
    return parse_bool(value, valuelen, &field_set, &_config.no_sub_mode);
}