// manager.c
//

#include "manager.h"

#include <assert.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <inttypes.h>

#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/wait.h>

#include <pktlab.h>

#include "config.h"
#include "console.h"
#include "util.h"
#include "worker.h"
#include "broker.h"
#include "xdescrpool.h"
#include "xdescr_util.h"
#include "comm.h"
#include "tls_util.h"
#include "cert.h"
#include "key.h"

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#ifndef SPEC_MAX
#define SPEC_MAX 256
#endif

#define SELECT_MAX pktlab_time_sec(2678400) // 31 days in sec

#define AUTHMODE PKTLAB_AUTH_AGENT
static struct pktlab_certificate *endpoint_agent_cert = NULL;
static struct pktlab_publickey **broker_troot_ls = NULL;
static size_t endpoint_trootnum = 0;
static struct pktlab_privatekey *privk = NULL;
static struct pktlab_publickey ** eopkey_ls = NULL;
static size_t eopkey_num = 0;
static struct pktctrl_keycert_store *store = NULL;

//
// INTERNAL TYPE DEFINITIONS
//

struct worker {
    struct worker *next;
    const char *spec;
    int fd;
    int pid;
    struct xdescr_node *node;
};

static struct worker *_workers = NULL;
static int _depfd = -1;
static pid_t _dep_pid = -1;
static pid_t _broker_pid = -1;
static char _confpath[PATH_MAX];
// static char _workerspec[SPEC_MAX];

//
// INTERNAL FUNCTION DECLARATIONS
//

static void create_deputy(int brokerfd, int closefd);
// static void create_worker(const char *spec);
static pid_t create_worker(struct xdescr_node *node, int depfd, int workerfd);
static void create_broker(int depfd, int closefd);

static void deputy_start(int mgrfd, int brokerfd);
// static void wait_workers(void);

static pid_t wait_child(int *status);
static int handle_child_exitval(pid_t pid, int status);
static int manager_install_sighandler(void);
static void manager_sighandler(int sig);

static int deputy_install_sighandler(void);
static void deputy_sighandler(int sig);

// static int sendspec(int depfd, const char *spec);
// static int recvspec(int mgrfd, char *buf, size_t bufsz);

// static int recvfd(int depfd);
// static int sendfd(int mgrfd, int outfd);

static int handle_worker_ctrlmsg(struct worker *w);

//
// EXPORTED FUNCTION DEFINITIONS
//

void manager_init(void) {
    trace("%s()", __func__);
    // ...
    // TODO:
    //   Implement init static data structures,
    //   and _workers, _depfd clean up functions
}

void manager_start(const char *confpath, const char *const spec) {
    int fds[2];
    int result;
    char * xdescr;
    uint_fast32_t conf_errlineno;
    int err = -128;
    trace("%s(confpath:%s,spec:%s)",
        __func__, safe_str(confpath), safe_str(spec));

    if (confpath != NULL)
        strncpy(_confpath, confpath, PATH_MAX - 1);

    result = socketpair(AF_LOCAL, SOCK_STREAM, 0, fds);
    perror_exit_if(result < 0, "socketpair");

    // the config file is parsed here
    // parsing addr/cert/keys are done by specific process broker/worker/deputy
    // when required
    config_initialize();
    if (config_read(_confpath, &conf_errlineno) < 0) {
        if (conf_errlineno == 0) {
            fatal("Failed to open config path");
        }

        fatal("Failed to parse config content "
              "(bad line %" PRIuFAST32 ")",
              conf_errlineno);
    } else if (!is_config_presence_correct(&err)) {
        debug("is_config_presence_correct returned err: %d", err);
        config_cleanup();
        fatal("Config option presence requirement not satisfied");
    }

    create_deputy(fds[0], fds[1]);

    if (!_config.no_sub_mode) {
        // sub mode, spawn broker process
        create_broker(fds[1], fds[0]);
        close(fds[0]);
        close(fds[1]);
    } else {
        // no sub mode, manager pass forged xdescrs to deputy directly
        close(fds[0]);

        result = get_xdescr_spec(spec, &xdescr);
        if (result < 0) {
            debug("get_xdescr_spec failed with %d", result);
            fatal("Cannot get xdescr for %s", spec);
        }

        if (sendxdescr(fds[1], xdescr, strlen(xdescr)) < 0) {
            debug("sendxdescr failed");
            fatal("Cannot send xdescr (%s) to deputy", xdescr);
        }

        close(fds[1]);
    }

    if (manager_install_sighandler()) {
        fatal("Failed to install signal handler");
    }
}

void manager_wait(void) {
    int rst, exit_stat = EXIT_SUCCESS;
    int child_stat, kill_pid;
    pid_t child_pid;

    trace("%s()", __func__);

    close(_depfd);

    child_pid = wait_child(&child_stat);
    exit_stat = handle_child_exitval(child_pid, child_stat);
    if (exit_stat == EXIT_FAILURE) {
        exit(EXIT_FAILURE);
    }

    if (!_config.no_sub_mode) { // only in sub mode are there two children
        // kill the broker process if deputy is terminated
        // deputy will auto exit if no new experiments are available
        kill_pid = child_pid == _dep_pid ? _broker_pid : _dep_pid;
        if (child_pid == _dep_pid) {
            rst = kill(kill_pid, SIGTERM);
            perror_exit_if(rst == -1, "kill");
        }

        child_pid = wait_child(&child_stat);
        if (child_pid != kill_pid) {
            warn("Unexpected child pid %d != %d", child_pid, kill_pid);
            exit(EXIT_FAILURE);
        }
        exit_stat = handle_child_exitval(child_pid, child_stat);
    }

    exit(exit_stat);
}

//
// INTERNAL FUNCTION DEFINITIONS
//

int handle_child_exitval(pid_t pid, int status) {
    char entity[16] = {};
    int rst = EXIT_SUCCESS;
    if (pid == -1) {
        warn("Deputy/broker cannot be waited");
        return EXIT_FAILURE;
    } else if (pid == -3) {
        warn("should not happen in POSIX.1-2001");
        return EXIT_FAILURE;
    }

    if (pid == _dep_pid) {
        strncpy(entity, "Deputy", sizeof(entity) - 1);
    } else if (pid == _broker_pid) {
        strncpy(entity, "Broker", sizeof(entity) - 1);
    } else {
        strncpy(entity, "Unknown", sizeof(entity) - 1);
        rst = EXIT_FAILURE;
    }

    if (WIFEXITED(status)) {
        info("%s process exit stat: %d", entity, WEXITSTATUS(status));
    } else {
        info("%s process exit stat: unknown", entity);
    }
    return rst;
}

void create_deputy(int brokerfd, int closefd) {
    int fds[2];
    int result;

    trace("%s()", __func__);

    result = socketpair(AF_LOCAL, SOCK_STREAM, 0, fds);
    perror_exit_if(result < 0, "socketpair");

    debug("socketpair() returned {%d,%d}", fds[0], fds[1]);

    if ((_dep_pid = fork())) {
        close(fds[0]);
        _depfd = fds[1];
        return;
    } else {
        close(fds[1]);
        close(closefd);
        if (deputy_install_sighandler()) {
            fatal("Failed to install deputy signal handler");
        }
        deputy_start(fds[0], brokerfd);
    }
}

pid_t create_worker(struct xdescr_node *node, int depfd, int workerfd) {
    pid_t pid = 0;
    struct worker *worker;
    if ((pid = fork())) {
        close(depfd);
        worker = safe_calloc(1, sizeof(*worker));
        worker->next = _workers;
        worker->fd = workerfd;
        worker->node = node;
        worker->spec = node->controller_addr;
        worker->pid = pid;
        _workers = worker;
        return pid;
    } else {
        struct sigaction sa;
        sa.sa_handler = SIG_DFL;
        if (sigaction(SIGTERM, &sa, NULL) != 0)
            fatal("Worker failed to install sighandler");
        return 0;
    }
}

// void create_worker(const char *spec) {
//     struct worker *worker;
//     int result;

//     trace("%s(spec:\"%s\")", __func__, spec);

//     result = sendspec(_depfd, spec);
//     if (result < 0)
//         exit(EXIT_FAILURE);

//     result = recvfd(_depfd);
//     if (result < 0)
//         exit(EXIT_FAILURE);

//     worker = safe_calloc(1, sizeof(struct worker));
//     worker->spec = spec;
//     worker->fd = result;
//     worker->next = _workers;
//     _workers = worker;
// }

void create_broker(int depfd, int closefd) {
    trace("%s %d", __func__, depfd);
    if ((_broker_pid = fork())) {
        return;
    } else {
        close(closefd);
        int result;
        result =
            load_cert(_config.endpoint_agent_cert_path, &endpoint_agent_cert);
        assert(result >= 0);

        result = ll_load_pubkeys(_config.broker_chain_troot_path_ll,
                                 &broker_troot_ls, &endpoint_trootnum);
        assert(result >= 0);

        result = load_privkey(_config.endpoint_privk_path, NULL, &privk);
        assert(result >= 0);

        store = pktctrl_create_keycert_store();
        assert(pktctrl_set_client_keycert(
                   store, &endpoint_agent_cert, 1, broker_troot_ls,
                   endpoint_trootnum) == PKTCTRL_SUCCESS);

        result = ll_load_pubkeys(_config.eop_pubk_path_ll, &eopkey_ls, &eopkey_num);
        broker_main(depfd, _config.broker_addr, NULL, privk, eopkey_ls, eopkey_num, store);
        trace("broker_main returns");
    }
}

void deputy_start(int mgrfd, int brokerfd) {
    int fds[2];
    int result;
    int nfds;
    fd_set rset, wset;
    struct timeval tv;
    static char xdescrbuf[4096];
    uint_fast32_t xdescrlen;
    int nhandle;
    struct xdescr_node *node;
    pktlab_time_t now, yield_hint = 0;
    bool brokerfd_eof = false;
    struct pktctrl_keycert_store *store = NULL;
    struct pktlab_certificate** endpoint_certls = NULL;
    uint_fast32_t endpoint_certnum;
    struct pktlab_publickey** endpoint_trootls = NULL;
    uint_fast32_t endpoint_trootnum;
    struct pktlab_privatekey * privk = NULL;

    trace("%s(mgrfd:%d)", __func__, mgrfd);
    debug("deputy PID is %d, PPID is %d", getpid(), getppid());

    worker_initialize();

    // create a keycert store for the worker
    if (ll_load_cert_chain(_config.controller_chain_path_ll, &endpoint_certls, &endpoint_certnum) < 0) {
        fatal("load controller cert chain failed");
    }

    if (ll_load_pubkeys(_config.eop_pubk_path_ll, &endpoint_trootls, &endpoint_trootnum) < 0) {
        fatal("load eop key failed");
    }

    store = pktctrl_create_keycert_store();
    if (pktctrl_set_client_keycert(
            store, endpoint_certls, endpoint_certnum,
            endpoint_trootls,
            endpoint_trootnum) != PKTCTRL_SUCCESS) {
        fatal("create worker keycert store failed");
    }

    if (load_privkey(_config.endpoint_privk_path, NULL, &privk)) {
        fatal("load endpoint private key failed");
    }

    // deputy manager mainloop
    while (true) {
        nfds = 0;
        FD_ZERO(&rset);
        FD_ZERO(&wset);
        now = pktlab_time_now();

        if (yield_hint > now) {
            pktlab_time_to_timeval(
                MIN(yield_hint - now, SELECT_MAX), &tv);
        } else {
            tv.tv_sec = 1;
            tv.tv_usec = 0;
        }

        if (!brokerfd_eof) {
            FD_SET(brokerfd, &rset);
            nfds = MAX(brokerfd, nfds);
        }

        for (struct worker *w = _workers; w != NULL; w = w->next) {
            FD_SET(w->fd, &rset);
            nfds = MAX(w->fd, nfds);
        }

        nfds++;
        nhandle = select(nfds, &rset, &wset, NULL, &tv);
        if (nhandle < 0) {
            fatal("select failed");
        }

        // check brokerfd for xdescr
        if (nhandle && FD_ISSET(brokerfd, &rset)) {
            nhandle--;

            result = recvxdescr(brokerfd, xdescrbuf, sizeof(xdescrbuf)-1);
            if (result < 0) {
                fatal("recvxdescr failed");
            } else if (result == 0) { // brokerfd EOF
                brokerfd_eof = true;
            } else {
                xdescrlen = result;
                xdescrbuf[xdescrlen] = '\0'; // for logging later

                if (!is_wellformed_xdescr(xdescrbuf, xdescrlen)) {
                    warn("Malformed xdescr: %s, ignoring", xdescrbuf);
                } else {
                    info("Got well-formed xdescr: %s", xdescrbuf);

                    node = xdescrnode_construct(xdescrbuf, xdescrlen);
                    if (node == NULL) { // null when invalid xdescr
                        warn("xdescr (%s) invalid, ignoring", xdescrbuf);
                    } else {
                        info("Inserting xdescr (%s) to xdescr pool", xdescrbuf);
                        xdescrpool_push(node);
                    }
                }
            }
        }

        // check child return (workerfd)
        // check yield from worker fd
        for (struct worker *w = _workers; w != NULL; w = w->next) {
            if (nhandle && FD_ISSET(w->fd, &rset)) {
                nhandle--;
                // read msg from worker
                result = handle_worker_ctrlmsg(w);
                assert(result == 0);
            }
        }

        node = xdescrpool_get(&yield_hint);
        if (node == NULL) {
            // no runnable experiment now or in the future
            if (yield_hint == UINT64_MAX && brokerfd_eof) {
                info("no runnable experiment now or in the future");
                exit(EXIT_SUCCESS);
            }

            continue;
        }

        // if higher priority
        if (_workers == NULL ||
            (_workers->node->exp_priority < node->exp_priority)) {
            if (_workers) {
                // IPC worker to susend
                // FIXME: race condition between suspend and finish/yield?
                // implement some ack from worker
                result = send_ctrlmsg(_workers->fd, CTRLMSG_TYPE_SUSPEND, 0);
                assert(result == 0);
                free(_workers);
                _workers = NULL;
            }
            // spawn new worker
            result = socketpair(AF_LOCAL, SOCK_STREAM, 0, fds);
            perror_exit_if(result < 0, "socketpair");
            if (create_worker(node, fds[0], fds[1])) {
                continue;
            } else
                break;
        }
    }

    // We get here if we're the worker process. Close descriptors we don't
    // need and enter the worker task.
    close(mgrfd);
    close(fds[1]);
    worker_main(
        fds[0], node->controller_addr,
        (node->snikey[0] == '\0') ? \
            NULL:node->snikey,
        store, privk);
    exit(EXIT_SUCCESS);
}

pid_t wait_child(int *status) {
    pid_t child_pid;

    // return val:
    // >= 0 for child return status
    // -1 for no child left
    // -2 for unknown child exit reason
    // -3 for none POSIX OS behavior

    while ((child_pid = wait(status)) < 0) {
        switch (errno) {
        case ECHILD: // No more child
            return -1;
        case EINTR:
            continue;
        default: // should not happen in POSIX.1-2001
            return -3;
        }
    }

    return child_pid;
}

// void wait_workers(void) {
//     int rst;

//     trace("%s()", __func__);

//     while ((rst = wait_child()) != -1) {
//         assert(rst >= -2);

//         if (rst == -2)
//             warn("Worker exited with unknown stat");
//         else
//             info("Worker exited with stat: %d", rst);
//     }
// }

// int sendspec(int depfd, const char *spec) {
//     size_t speclen;
//     ssize_t size;

//     trace("%s(depfd:%d,spec:\"%s\")", __func__, depfd, spec);

//     speclen = strlen(spec);
//     size = write(depfd, spec, speclen + 1);

//     if (size < 0) {
//         if (errno == EPIPE)
//             return -1;
//         perror(__func__);
//         exit(EXIT_FAILURE);
//     }

//     return speclen;
// }

// int recvspec(int mgrfd, char *buf, size_t bufsz) {
//     size_t len = 0;
//     char *sptr;
//     ssize_t rdlen;

//     trace("%s(mgrfd:%d,bufsz:%zu)", __func__, mgrfd, bufsz);

//     // Receive a null-terminaled string containing spec into buf. We need to
//     // loop because a single message on a stream may be split into multiple
//     // reads. Since we only send one string at a time, there is no data to
//     // buffer for the next read after the null terminator.

//     do {
//         rdlen = read(mgrfd, buf + len, bufsz - len);

//         if (rdlen <= 0) {
//             if (rdlen == 0)
//                 return 0;
//             perror(__func__);
//             exit(EXIT_FAILURE);
//         }

//         sptr = memchr(buf + len, '\0', rdlen);

//         if (sptr != NULL) {
//             if (sptr + 1 != buf + len + rdlen) {
//                 perror("recvspec: unexpected data");
//                 exit(EXIT_FAILURE);
//             } else
//                 return len + rdlen;
//         }
//     } while ((len += rdlen) < bufsz);

//     // We get here if the spec string is larger than bufsz. This should never
//     // happen, so we panic.

//     perror("recvspec: buffer overflow");
//     exit(EXIT_FAILURE);
// }

// int recvfd(int depfd) {
//     char bytebuf;
//     struct iovec iov = {.iov_base = &bytebuf, .iov_len = sizeof(bytebuf)};
//     char cmsgbuf[sizeof(struct cmsghdr) + sizeof(int)]; // XXX need to align
//     struct cmsghdr *cmsgptr = (void *)cmsgbuf;
//     struct msghdr msg;
//     ssize_t size;
//     int infd;

//     memset(&msg, 0, sizeof(msg));
//     memset(cmsgbuf, 0, sizeof(cmsgbuf));
//     msg.msg_control = cmsgptr;
//     msg.msg_controllen = sizeof(cmsgbuf);
//     msg.msg_iov = &iov;
//     msg.msg_iovlen = 1;

//     // We expect a 1-byte message containing a descriptor.

//     size = recvmsg(depfd, &msg, 0);

//     if (size <= 0) {
//         if (size == 0)
//             return -1;
//         perror(__func__);
//         exit(EXIT_FAILURE);
//     }

//     memcpy(&infd, CMSG_DATA(cmsgptr), sizeof(int));
//     return infd;
// }

// int sendfd(int mgrfd, int outfd) {
//     char bytebuf = '\0';
//     struct iovec iov = {.iov_base = &bytebuf, .iov_len = sizeof(bytebuf)};
//     char cmsgbuf[sizeof(struct cmsghdr) + sizeof(int)]; // XXX need to align
//     struct cmsghdr *cmsgptr = (void *)cmsgbuf;
//     struct msghdr msg;
//     size_t size;

//     memset(&msg, 0, sizeof(msg));
//     msg.msg_control = cmsgptr;
//     msg.msg_controllen = sizeof(cmsgbuf);
//     msg.msg_iov = &iov;
//     msg.msg_iovlen = 1;

//     memset(cmsgbuf, 0, sizeof(cmsgbuf));
//     memcpy(CMSG_DATA(cmsgptr), &outfd, sizeof(int));
//     cmsgptr->cmsg_len = sizeof(cmsgbuf);
//     cmsgptr->cmsg_level = SOL_SOCKET;
//     cmsgptr->cmsg_type = SCM_RIGHTS;

//     size = sendmsg(mgrfd, &msg, MSG_EOR);

//     if (size < 0) {
//         if (errno == EPIPE)
//             return -1;
//         perror(__func__);
//         exit(EXIT_FAILURE);
//     } else
//         return outfd;
// }

int handle_worker_ctrlmsg(struct worker *w) {
    struct ctrlmsg msg;
    int rst;
    int status;
    trace("%s workerfd(%d)", __func__, w->fd);
    rst = recv_ctrlmsg(w->fd, &msg);
    if (rst != 0)
        return -1;
    // FIXME: accomodate more workers
    assert(_workers == w);
    if (msg.type == CTRLMSG_TYPE_YIELD) {
        w->node->yield_hint = msg.yield.time;
        _workers = NULL;
        free(w);
    } else if (msg.type == CTRLMSG_TYPE_FINISH) {
        // remove finished experiment from the pool
        xdescrpool_remove(w->node);
        xdescrnode_destroy(w->node);
        _workers = NULL;
        // cleanup the child process
        rst = waitpid(w->pid, &status, 0);
        info("Worker %d returns %d", w->pid, status);
        free(w);

    } else {
        warn("unexpected ctrlmsg type");
    }
    return 0;
}

int manager_install_sighandler(void) {
    struct sigaction sa;
    sa.sa_handler = &manager_sighandler;
    return (sigaction(SIGTERM, &sa, NULL) == 0 &&
            sigaction(SIGINT, &sa, NULL) == 0)
               ? 0
               : -1;
}

void manager_sighandler(int sig) {
    info("Manager receives signal %s", strsignal(sig));
    if (_dep_pid != -1) {
        kill(_dep_pid, SIGTERM);
    }
    if (_broker_pid != -1) {
        kill(_broker_pid, SIGTERM);
    }
    manager_wait();
}

int deputy_install_sighandler(void) {
    struct sigaction sa;
    sa.sa_handler = &deputy_sighandler;
    return (sigaction(SIGTERM, &sa, NULL) == 0 &&
            sigaction(SIGINT, &sa, NULL) == 0)
               ? 0
               : -1;
}

void deputy_sighandler(int sig) {
    info("Deputy receives signal %s", strsignal(sig));
    int child_stat;
    pid_t child_pid;
    char status_buf[16];
    // kill all running workers
    for (struct worker *w = _workers; w != NULL; w = w->next) {
        kill(w->pid, SIGTERM);
    }

    // cleanup the killed workers
    while ((child_pid = wait(&child_stat)) != -1) {
        if (child_pid < 0) {
            if (child_pid == EINTR) {
                continue;
            } else if (child_pid == -3) {
                warn("should not happen in POSIX.1-2001");
                exit(EXIT_FAILURE);
            }
        }
        snprintf(status_buf, sizeof(status_buf) - 1, "%d",
                 WEXITSTATUS(child_stat));
        info("Child %d exits with status %s", child_pid,
             WIFEXITED(child_stat) ? status_buf : "unknown");
    }
    exit(EXIT_SUCCESS);
}