// raw.h
// General raw socket module interface
//

// todo: update api

#ifndef _RAW_H_
#define _RAW_H_

#include <sys/types.h>

#include <pktlab.h>

//
// EXPORTED MACROS
//

#define RAW_PROG_INITIALIZER {.prog_fd = -1}

//
// EXPORTED CONSTANTS
//

enum raw_socket_state {
    ST_RAWPREINIT,
    ST_RAWINIT,
    ST_RAWIDLE,
    ST_RAWERROR,
    ST_RAWUNKNOWN
};

// relies on enum pktlab_nsend_err, following the order specified with BADPKT being max
// should not contain too much entry s.t. PKTLAB_NSEND_UNKFAULT is used twice
enum raw_error {
    RAW_SUCCESS,
    RAW_NORES,
    RAW_BADPKT,
    RAW_INVAL,
    RAW_TRYAGAIN,
    RAW_PERM,
    RAW_BADFILT,
    RAW_UNKFAULT,
};

//
// EXPORTED TYPE DEFINITIONS
//

// the fields are exposed for ease of use
// one should however never modify them directly
struct raw_socket_obj {
    int af;
    int rfd;
    int wfd;
    int ifindex;
    enum raw_socket_state st;
};

struct raw_prog {
    uint8_t * progptr;
    uint_fast32_t proglen;
    bool is_ebpf;
    int prog_fd; // for caching loaded programs; used only in filter_linux
    struct raw_prog * next;
};

//
// EXPORTED FUNCTION DECLARATIONS
//

extern struct raw_socket_obj * create_raw_socket_obj(void);
extern int raw_socket_setup_v4 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj);
extern int raw_socket_setup_v6 (
    const uint8_t * intfaddr,
    struct raw_prog * cert_prog_ll,
    struct raw_socket_obj * obj);
extern void raw_socket_close(struct raw_socket_obj * obj);
extern ssize_t raw_socket_send (
    const void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);
extern ssize_t raw_socket_recv (
    void * buf, size_t len, int flags,
    struct raw_socket_obj * obj, enum raw_error * err);
extern int raw_socket_set_recvfilter (
    struct raw_prog * filter,
    struct raw_socket_obj * obj);
extern void raw_v4_hdrinfo_read (
    const struct pktlab_vmem_region * restrict rgn,
    uint_fast32_t off, uint_fast32_t len, void * buf);
extern void raw_v6_hdrinfo_read (
    const struct pktlab_vmem_region * restrict rgn,
    uint_fast32_t off, uint_fast32_t len, void * buf);
static inline const char * raw_strerror(enum raw_error err);

// todo: update doc below
// look how I massacred my boy ...

// All functions return -1 for error notification, actual error stored in "err" var.
// If function call successful, err should be set to RAW_SUCCESS.
//

// Initialize raw socket module, must be called before raw_start
// Return:
//      0 for success
// Error:
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern int raw_init(enum raw_error *err);

// Start raw socket module, called after raw_init
// Parameters:
//      af: raw socket address family
//      fds[0]: recvfd
//      fds[1]: sendfd
// Return:
//      0 for success
// Error:
//      RAW_INVAL for bad address family
//      RAW_AFNOSUPPORT for system not supporting the specified address family
//          (need stop and reinit)
//      RAW_PERM for not having the permission to open raw sockets (need stop and reinit)
//      RAW_NORES for not enough system resources to start module (need stop and reinit)
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern int raw_start(int af, int fds[2], enum raw_error *err);

// Raw socket send traffic
// Parameters:
//      buf: packet content
//      len: packet length in bytes
//      flags: flags for sendto function call
// Return:
//      >= 0 sent bytes when successful
// Error:
//      RAW_TRYAGAIN for non-blocking send not possible or interrupted
//      RAW_BADPKT for bad packets (rejected by system)
//      RAW_INVAL for invalid argument including flags
//      RAW_NORES for machine resources not available (need stop and reinit)
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern ssize_t raw_send (
//    const void * buf, size_t len, int flags,
//    struct raw_socket_obj * obj, enum raw_error * err);

// Raw socket recv traffic
// Parameters:
//      buf: buffer for packet content
//      len: buffer length in bytes
//      flags: flags for recvfrom function call
// Return:
//      >= 0 received bytes when successful
// Error:
//      RAW_TRYAGAIN for non-blocking recv not possible or interrupted
//      RAW_INVAL for invalid argument including flags
//      RAW_NORES for machine resources not available (need stop and reinit)
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern ssize_t raw_recv (
//    void * buf, size_t len, int flags,
//    struct raw_socket_obj * obj, enum raw_error * err);

// Stop raw socket module
// Return:
//      0 for success
// Error:
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern int raw_stop(enum raw_error *err);

// Set raw recv filter, set all drop filter if filter is NULL
// Parameters:
//      filter: pointer to bpf filter code
//      len: bpf filter byte length
// Return:
//      0 for success
// Error:
//      RAW_BADFILT for bad supplied bpf filter
//      RAW_UNKFAULT for unknown error (need stop and reinit)
//extern int raw_setfilter(void * filter, size_t len, enum raw_error *err);


// Module utility functions

//static inline const char * raw_strerror(enum raw_error err);

// Memory read handlers for exporting raw module info

// Exports information on how may the raw socket
// modify the ip headers before sending the raw packet out.
// Information exported via a uint8_t array with each
// element corresponding to a header field.
// The meaning of the uint8_t element value is as follows:
//      PKTLAB_RAWHDR_NOCHG:
//          The field will not be changed.
//      PKTLAB_RAWHDR_CHG:
//          The field will always be changed.
//      PKTLAB_RAWHDR_0FILL:
//          The field will be changed when the field is set to all 0.
// The values are defined as enum pktlab_rawhdr_info in pktlab.h

// A 12 element uint8_t array corresponding to
// the 12 header fields in the IPv4 header.
// Note that we don't export info about IPv4 options.
//
// Offset, header field correspondence:
//      OFF HEADER
//      --- -----------
//        0 Version
//        1 IHL
//        2 Type of Service (TOS)
//        3 Total Length
//        4 Identification
//        5 Flags
//        6 Fragment offset
//        7 Time to Live (TTL)
//        8 Protocol
//        9 Header Checksum
//       10 Source Address
//       11 Destination Address
//extern void raw_v4_hdrinfo_read (
//	const struct pktlab_vmem_region * restrict rgn,
//	uint_fast32_t off, uint_fast32_t len, void * buf);

// An 8 element uint8_t array corresponding to
// the 8 header fields in the IPv6 header.
// Note that we don't export info about IPv6 options.
//
// Offset, header field correspondence:
//      OFF HEADER
//      --- -----------
//        0 Version
//        1 Traffic Class
//        2 Flow Label
//        3 Payload Length
//        4 Next Header
//        5 Hop Limit
//        6 Source Address
//        7 Destination Address
//extern void raw_v6_hdrinfo_read (
//	const struct pktlab_vmem_region * restrict rgn,
//	uint_fast32_t off, uint_fast32_t len, void * buf);

//
// SHARED FUNCTION DEFINITION
//

static inline const char * raw_strerror(enum raw_error err) {
    static const char * const errmsgs[] = {
        [RAW_SUCCESS] = "Operation successful",
        [RAW_INVAL] = "Bad function arguments",
        [RAW_PERM] = "No permission for this operation",
        [RAW_TRYAGAIN] = "Temporarily cannot perform operation",
        [RAW_BADPKT] = "Bad packet supplied",
        [RAW_BADFILT] = "Bad filter supplied",
        [RAW_NORES] = "System out of resources",
        [RAW_UNKFAULT] = "Unknown error"
    };

    const char * errmsg = NULL;
    switch(err) {
    case RAW_SUCCESS:
    case RAW_INVAL:
    case RAW_PERM:
    case RAW_TRYAGAIN:
    case RAW_BADPKT:
    case RAW_BADFILT:
    case RAW_NORES:
        errmsg = errmsgs[err];
        break;
    default:
        errmsg = errmsgs[RAW_UNKFAULT];
    }

    return errmsg;
}

#endif